def apply_metaclass_to_cls(cls, metaclass):
    # Create a new class using the metaclass, with the same name, bases, and attributes as the original class
    new_class = metaclass(cls.__name__, cls.__bases__, dict(cls.__dict__))
    return new_class



from typing import Optional, Any, Literal
from pydantic import Field
from swarmauri_base.messages.MessageBase import MessageBase
from swarmauri_core.ComponentBase import ComponentBase

@ComponentBase.register_type(MessageBase, 'SystemMessage')
class SystemMessage(MessageBase):
    content: str
    role: str = Field(default='system')
    type: Literal['SystemMessage'] = 'SystemMessage'


from typing import List, Literal, Dict
from pydantic import Field
from swarmauri_standard.tools.Parameter import Parameter
from swarmauri_base.tools.ToolBase import ToolBase
from swarmauri_core.ComponentBase import ComponentBase

@ComponentBase.register_type(ToolBase, 'TestTool')
class TestTool(ToolBase):
    version: str = "1.0.0"

    # Define the parameters required by the tool
    parameters: List[Parameter] = Field(
        default_factory=lambda: [
            Parameter(
                name="program",
                type="string",
                description="The program that the user wants to open ('notepad' or 'calc' or 'mspaint')",
                required=True,
                enum=["notepad", "calc", "mspaint"],
            )
        ]
    )
    name: str = "TestTool"
    description: str = "This opens a program based on the user's request."
    type: Literal["TestTool"] = "TestTool"

    def __call__(self, program) -> Dict[str, str]:
        # sp.check_output(program)
        # Here you would implement the actual logic for fetching the weather information.
        # For demonstration, let's just return the parameters as a string.
        return {"program": f"Program Opened: {program}"}




# from swarmauri_standard.utils._lazy_import import _lazy_import

# # List of pipeline names (file names without the ".py" extension) and corresponding class names
# pipeline_files = [
#     ("swarmauri_standard.pipelines.Pipeline", "Pipeline"),
# ]

# # Lazy loading of pipeline classes, storing them in variables
# for module_name, class_name in pipeline_files:
#     globals()[class_name] = _lazy_import(module_name, class_name)

# # Adding the lazy-loaded pipeline classes to __all__
# __all__ = [class_name for _, class_name in pipeline_files]



from pydantic import PrivateAttr
import asyncio
import httpx
from typing import Dict, List, Literal, Optional
from swarmauri_base.image_gens.ImageGenBase import ImageGenBase
from swarmauri_standard.utils.retry_decorator import retry_on_status_codes
from swarmauri_core.ComponentBase import ComponentBase

@ComponentBase.register_type(ImageGenBase, 'OpenAIImgGenModel')
class OpenAIImgGenModel(ImageGenBase):
    """
    OpenAIImgGenModel is a class for generating images using OpenAI's DALL-E models.

    Attributes:
        api_key (str): The API key for authenticating with the OpenAI API.
        allowed_models (List[str]): List of allowed model names.
        name (str): The name of the model to use.
        type (Literal["OpenAIImgGenModel"]): The type of the model.

    Provider Resources: https://platform.openai.com/docs/api-reference/images/generate
    """

    api_key: str
    allowed_models: List[str] = ["dall-e-2", "dall-e-3"]
    name: str = "dall-e-3"
    type: Literal["OpenAIImgGenModel"] = "OpenAIImgGenModel"
    _BASE_URL: str = PrivateAttr(default="https://api.openai.com/v1/images/generations")
    _headers: Dict[str, str] = PrivateAttr(default=None)

    def __init__(self, **kwargs) -> None:
        """
        Initialize the GroqAIAudio class with the provided data.

        Args:
            **data: Arbitrary keyword arguments containing initialization data.
        """
        super().__init__(**kwargs)
        self._headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }

    @retry_on_status_codes((429, 529), max_retries=1)
    def generate_image(
        self,
        prompt: str,
        size: str = "1024x1024",
        quality: str = "standard",
        n: int = 1,
        style: Optional[str] = None,
    ) -> List[str]:
        """
        Generate images using the OpenAI DALL-E model synchronously.

        Parameters:
        - prompt (str): The prompt to generate images from.
        - size (str): Size of the generated images.
        - quality (str): Quality of the generated images.
        - n (int): Number of images to generate.
        - style (str): Optional style of the generated images.

        Returns:
        - List of URLs of the generated images.
        """
        if self.name == "dall-e-3" and n > 1:
            raise ValueError("DALL-E 3 only supports generating 1 image at a time.")

        payload = {
            "model": self.name,
            "prompt": prompt,
            "size": size,
            "quality": quality,
            "n": n,
        }

        if style and self.name == "dall-e-3":
            payload["style"] = style

        try:
            with httpx.Client(timeout=30.0) as client:
                response = client.post(
                    self._BASE_URL, headers=self._headers, json=payload
                )
                response.raise_for_status()
                return [image["url"] for image in response.json().get("data", [])]
        except httpx.HTTPStatusError as e:
            raise RuntimeError(f"Image generation failed: {e}")

    @retry_on_status_codes((429, 529), max_retries=1)
    async def agenerate_image(
        self,
        prompt: str,
        size: str = "1024x1024",
        quality: str = "standard",
        n: int = 1,
        style: Optional[str] = None,
    ) -> List[str]:
        """
        Generate images using the OpenAI DALL-E model asynchronously.

        Parameters:
        - prompt (str): The prompt to generate images from.
        - size (str): Size of the generated images.
        - quality (str): Quality of the generated images.
        - n (int): Number of images to generate.
        - style (str): Optional style of the generated images.

        Returns:
        - List of URLs of the generated images.
        """
        if self.name == "dall-e-3" and n > 1:
            raise ValueError("DALL-E 3 only supports generating 1 image at a time.")

        payload = {
            "model": self.name,
            "prompt": prompt,
            "size": size,
            "quality": quality,
            "n": n,
        }

        if style and self.name == "dall-e-3":
            payload["style"] = style

        try:
            async with httpx.AsyncClient(timeout=30.0) as client:
                response = await client.post(
                    self._BASE_URL, headers=self._headers, json=payload
                )
                response.raise_for_status()
                return [image["url"] for image in response.json().get("data", [])]
        except httpx.HTTPStatusError as e:
            raise RuntimeError(f"Image generation failed: {e}")

    def batch(
        self,
        prompts: List[str],
        size: str = "1024x1024",
        quality: str = "standard",
        n: int = 1,
        style: Optional[str] = None,
    ) -> List[List[str]]:
        """
        Synchronously process multiple prompts for image generation.

        Parameters:
        - prompts (List[str]): List of prompts.
        - size (str): Size of the generated images.
        - quality (str): Quality of the generated images.
        - n (int): Number of images to generate.
        - style (str): Optional style of the generated images.

        Returns:
        - List of lists of URLs of the generated images.
        """
        return [
            self.generate_image(prompt, size=size, quality=quality, n=n, style=style)
            for prompt in prompts
        ]

    async def abatch(
        self,
        prompts: List[str],
        size: str = "1024x1024",
        quality: str = "standard",
        n: int = 1,
        style: Optional[str] = None,
        max_concurrent: int = 5,
    ) -> List[List[str]]:
        """
        Asynchronously process multiple prompts for image generation with controlled concurrency.

        Parameters:
        - prompts (List[str]): List of prompts.
        - size (str): Size of the generated images.
        - quality (str): Quality of the generated images.
        - n (int): Number of images to generate.
        - style (str): Optional style of the generated images.
        - max_concurrent (int): Maximum number of concurrent requests.

        Returns:
        - List of lists of URLs of the generated images.
        """
        semaphore = asyncio.Semaphore(max_concurrent)

        async def process_prompt(prompt) -> List[str]:
            async with semaphore:
                return await self.agenerate_image(
                    prompt, size=size, quality=quality, n=n, style=style
                )

        tasks = [process_prompt(prompt) for prompt in prompts]
        return await asyncio.gather(*tasks)



# from swarmauri_standard.utils._lazy_import import _lazy_import

# # List of schema_converters names (file names without the ".py" extension) and corresponding class names
# schema_converters_files = [
#     (
#         "swarmauri_standard.schema_converters.AnthropicSchemaConverter",
#         "AnthropicSchemaConverter",
#     ),
#     (
#         "swarmauri_standard.schema_converters.CohereSchemaConverter",
#         "CohereSchemaConverter",
#     ),
#     (
#         "swarmauri_standard.schema_converters.GeminiSchemaConverter",
#         "GeminiSchemaConverter",
#     ),
#     ("swarmauri_standard.schema_converters.GroqSchemaConverter", "GroqSchemaConverter"),
#     (
#         "swarmauri_standard.schema_converters.MistralSchemaConverter",
#         "MistralSchemaConverter",
#     ),
#     (
#         "swarmauri_standard.schema_converters.OpenAISchemaConverter",
#         "OpenAISchemaConverter",
#     ),
#     (
#         "swarmauri_standard.schema_converters.ShuttleAISchemaConverter",
#         "ShuttleAISchemaConverter",
#     ),
# ]

# # Lazy loading of schema_converters classes, storing them in variables
# for module_name, class_name in schema_converters_files:
#     globals()[class_name] = _lazy_import(module_name, class_name)

# # Adding the lazy-loaded schema_converters classes to __all__
# __all__ = [class_name for _, class_name in schema_converters_files]



from typing import Literal
from swarmauri_base.chains.ChainStepBase import ChainStepBase
from swarmauri_core.ComponentBase import ComponentBase

@ComponentBase.register_type(ChainStepBase, 'ChainStep')
class ChainStep(ChainStepBase):
    """
    Represents a single step within an execution chain.
    """
    type: Literal['ChainStep'] = 'ChainStep'



import httpx
import asyncio
import time
from typing import List, Literal, Optional, Dict
from pydantic import Field, PrivateAttr
from swarmauri_standard.utils.retry_decorator import retry_on_status_codes
from swarmauri_base.image_gens.ImageGenBase import ImageGenBase
from swarmauri_core.ComponentBase import ComponentBase

@ComponentBase.register_type(ImageGenBase, 'FalAIImgGenModel')
class FalAIImgGenModel(ImageGenBase):
    """
    A model class for generating images from text using FluxPro's image generation model,
    provided by FalAI. This class uses a queue-based API to handle image generation requests.

    Attributes:
        allowed_models (List[str]): List of valid model names for image generation.
        api_key (str): The API key for authenticating requests with the FalAI service.
        model_name (str): The name of the model used for image generation.
        type (Literal): The model type, fixed as "FalAIImgGenModel".
        max_retries (int): The maximum number of retries for polling request status.
        retry_delay (float): Delay in seconds between status check retries.
    """

    _BASE_URL: str = PrivateAttr("https://queue.fal.run")
    _client: httpx.Client = PrivateAttr()
    _async_client: Optional[httpx.AsyncClient] = PrivateAttr(default=None)

    allowed_models: List[str] = [
        "fal-ai/flux-pro",
    ]
    api_key: str = Field(default=None)
    name: str = Field(default="fal-ai/flux-pro")
    type: Literal["FalAIImgGenModel"] = "FalAIImgGenModel"
    max_retries: int = Field(default=60)  # Maximum number of status check retries
    retry_delay: float = Field(default=1.0)  # Delay between status checks in seconds

    def __init__(self, **kwargs):
        """
        Initializes the model with the specified API key and model name.

        Args:
            **data: Configuration parameters for the model.

        Raises:
            ValueError: If an invalid model name is provided.
        """
        super().__init__(**kwargs)
        self._headers = {
            "Content-Type": "application/json",
            "Authorization": f"Key {self.api_key}",
        }
        self._client = httpx.Client(headers=self._headers, timeout=30)

    async def _get_async_client(self) -> httpx.AsyncClient:
        """
        Get or create the async client.

        Returns:
            httpx.AsyncClient: The async HTTP client instance.
        """
        if self._async_client is None or self._async_client.is_closed:
            self._async_client = httpx.AsyncClient(headers=self._headers, timeout=30)
        return self._async_client

    async def _close_async_client(self):
        """
        Safely close the async client if it exists and is open.
        """
        if self._async_client is not None and not self._async_client.is_closed:
            await self._async_client.aclose()
            self._async_client = None

    def _create_request_payload(self, prompt: str, **kwargs) -> dict:
        """
        Creates a payload for the image generation request.

        Args:
            prompt (str): The text prompt for image generation.
            **kwargs: Additional parameters for the request.

        Returns:
            dict: The request payload.
        """
        return {"prompt": prompt, **kwargs}

    @retry_on_status_codes((429, 529), max_retries=1)
    def _send_request(self, prompt: str, **kwargs) -> Dict:
        """
        Sends an image generation request to the queue and returns the request ID.

        Args:
            prompt (str): The text prompt for image generation.
            **kwargs: Additional parameters for the request.

        Returns:
            Dict: The response containing the request ID.
        """
        url = f"{self._BASE_URL}/{self.name}"
        payload = self._create_request_payload(prompt, **kwargs)

        response = self._client.post(url, json=payload)
        response.raise_for_status()
        return response.json()

    @retry_on_status_codes((429, 529), max_retries=1)
    def _check_status(self, request_id: str) -> Dict:
        """
        Checks the status of a queued image generation request.

        Args:
            request_id (str): The ID of the request.

        Returns:
            Dict: The response containing the request status.
        """
        url = f"{self._BASE_URL}/{self.name}/requests/{request_id}/status"
        response = self._client.get(url, params={"logs": 1})
        response.raise_for_status()
        return response.json()

    @retry_on_status_codes((429, 529), max_retries=1)
    def _get_result(self, request_id: str) -> Dict:
        """
        Retrieves the final result of a completed request.

        Args:
            request_id (str): The ID of the completed request.

        Returns:
            Dict: The response containing the generated image URL.
        """
        url = f"{self._BASE_URL}/{self.name}/requests/{request_id}"
        response = self._client.get(url)
        response.raise_for_status()
        return response.json()

    @retry_on_status_codes((429, 529), max_retries=1)
    async def _async_send_request(self, prompt: str, **kwargs) -> Dict:
        """
        Asynchronously sends an image generation request to the queue.

        Args:
            prompt (str): The text prompt for image generation.
            **kwargs: Additional parameters for the request.

        Returns:
            Dict: The response containing the request ID.
        """
        client = await self._get_async_client()
        url = f"{self._BASE_URL}/{self.name}"
        payload = self._create_request_payload(prompt, **kwargs)

        response = await client.post(url, json=payload)
        response.raise_for_status()
        return response.json()

    @retry_on_status_codes((429, 529), max_retries=1)
    async def _async_check_status(self, request_id: str) -> Dict:
        """
        Asynchronously checks the status of a queued request.

        Args:
            request_id (str): The ID of the request.

        Returns:
            Dict: The response containing the request status.
        """
        client = await self._get_async_client()
        url = f"{self._BASE_URL}/{self.name}/requests/{request_id}/status"
        response = await client.get(url, params={"logs": 1})
        response.raise_for_status()
        return response.json()

    @retry_on_status_codes((429, 529), max_retries=1)
    async def _async_get_result(self, request_id: str) -> Dict:
        """
        Asynchronously retrieves the final result of a completed request.

        Args:
            request_id (str): The ID of the completed request.

        Returns:
            Dict: The response containing the generated image URL.
        """
        client = await self._get_async_client()
        url = f"{self._BASE_URL}/{self.name}/requests/{request_id}"
        response = await client.get(url)
        response.raise_for_status()
        return response.json()

    def _wait_for_completion(self, request_id: str) -> Dict:
        """
        Waits for a request to complete, polling the status endpoint.

        Args:
            request_id (str): The ID of the request.

        Returns:
            Dict: The final response containing the generated image URL.

        Raises:
            TimeoutError: If the request does not complete within the retry limit.
        """
        for _ in range(self.max_retries):
            status_data = self._check_status(request_id)
            if status_data["status"] == "COMPLETED":
                return self._get_result(request_id)
            elif status_data["status"] in ["IN_QUEUE", "IN_PROGRESS"]:
                time.sleep(self.retry_delay)
            else:
                raise RuntimeError(f"Unexpected status: {status_data}")

        raise TimeoutError(
            f"Request {request_id} did not complete within the timeout period"
        )

    async def _async_wait_for_completion(self, request_id: str) -> Dict:
        """
        Asynchronously waits for a request to complete, polling the status endpoint.

        Args:
            request_id (str): The ID of the request.

        Returns:
            Dict: The final response containing the generated image URL.

        Raises:
            TimeoutError: If the request does not complete within the retry limit.
        """
        for _ in range(self.max_retries):
            status_data = await self._async_check_status(request_id)
            if status_data["status"] == "COMPLETED":
                return await self._async_get_result(request_id)
            elif status_data["status"] in ["IN_QUEUE", "IN_PROGRESS"]:
                await asyncio.sleep(self.retry_delay)
            else:
                raise RuntimeError(f"Unexpected status: {status_data}")

        raise TimeoutError(
            f"Request {request_id} did not complete within the timeout period"
        )

    def generate_image(self, prompt: str, **kwargs) -> str:
        """
        Generates an image based on the prompt and returns the image URL.

        Args:
            prompt (str): The text prompt for image generation.
            **kwargs: Additional parameters for the request.

        Returns:
            str: The URL of the generated image.
        """
        initial_response = self._send_request(prompt, **kwargs)
        request_id = initial_response["request_id"]
        final_response = self._wait_for_completion(request_id)
        return final_response["images"][0]["url"]

    async def agenerate_image(self, prompt: str, **kwargs) -> str:
        """
        Asynchronously generates an image based on the prompt and returns the image URL.

        Args:
            prompt (str): The text prompt for image generation
            **kwargs: Additional parameters to pass to the API

        Returns:
            str: The URL of the generated image
        """
        try:
            initial_response = await self._async_send_request(prompt, **kwargs)
            request_id = initial_response["request_id"]
            final_response = await self._async_wait_for_completion(request_id)
            return final_response["images"][0]["url"]
        finally:
            await self._close_async_client()

    def batch(self, prompts: List[str], **kwargs) -> List[str]:
        """
        Generates images for a batch of prompts.

        Args:
            prompts (List[str]): List of text prompts
            **kwargs: Additional parameters to pass to the API

        Returns:
            List[str]: List of image URLs
        """
        return [self.generate_image(prompt, **kwargs) for prompt in prompts]

    async def abatch(
        self, prompts: List[str], max_concurrent: int = 5, **kwargs
    ) -> List[str]:
        """
        Asynchronously generates images for a batch of prompts.

        Args:
            prompts (List[str]): List of text prompts
            max_concurrent (int): Maximum number of concurrent requests
            **kwargs: Additional parameters to pass to the API

        Returns:
            List[str]: List of image URLs
        """
        try:
            semaphore = asyncio.Semaphore(max_concurrent)

            async def process_prompt(prompt):
                async with semaphore:
                    initial_response = await self._async_send_request(prompt, **kwargs)
                    request_id = initial_response["request_id"]
                    final_response = await self._async_wait_for_completion(request_id)
                    return final_response["response"]["images"][0]["url"]

            tasks = [process_prompt(prompt) for prompt in prompts]
            return await asyncio.gather(*tasks)
        finally:
            await self._close_async_client()

    def __del__(self):
        """Cleanup method to close HTTP clients."""
        self._client.close()



from typing import List, Union, Any, Literal
import joblib
from pydantic import PrivateAttr
from sklearn.feature_extraction.text import TfidfVectorizer as SklearnTfidfVectorizer

from swarmauri_base.embeddings.EmbeddingBase import EmbeddingBase
from swarmauri_standard.vectors.Vector import Vector
from swarmauri_core.ComponentBase import ComponentBase

@ComponentBase.register_type(EmbeddingBase, 'TfidfEmbedding')
class TfidfEmbedding(EmbeddingBase):
    _model = PrivateAttr()
    _fit_matrix = PrivateAttr()
    type: Literal["TfidfEmbedding"] = "TfidfEmbedding"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._model = SklearnTfidfVectorizer()

    def extract_features(self):
        return self._model.get_feature_names_out().tolist()

    def fit(self, documents: List[str]) -> None:
        self._fit_matrix = self._model.fit_transform(documents)

    def fit_transform(self, documents: List[str]) -> List[Vector]:
        self._fit_matrix = self._model.fit_transform(documents)
        # Convert the sparse matrix rows into Vector instances
        vectors = [
            Vector(value=vector.toarray().flatten()) for vector in self._fit_matrix
        ]
        return vectors

    def transform(self, data: Union[str, Any], documents: List[str]) -> List[Vector]:
        raise NotImplementedError("Transform not implemented on TFIDFVectorizer.")

    def infer_vector(self, data: str, documents: List[str]) -> Vector:
        documents.append(data)
        tmp_tfidf_matrix = self.fit_transform(documents)
        query_vector = tmp_tfidf_matrix[-1]
        return query_vector

    def save_model(self, path: str) -> None:
        """
        Saves the TF-IDF model to the specified path using joblib.
        """
        joblib.dump(self._model, path)

    def load_model(self, path: str) -> None:
        """
        Loads a TF-IDF model from the specified path using joblib.
        """
        self._model = joblib.load(path)



from typing import Any, Optional, Dict, Literal

from swarmauri_standard.conversations.MaxSystemContextConversation import MaxSystemContextConversation
from swarmauri_standard.messages.HumanMessage import HumanMessage

from swarmauri_base.llms.LLMBase import LLMBase
from swarmauri_base.agents.AgentBase import AgentBase
from swarmauri_core.ComponentBase import SubclassUnion, ComponentBase

@ComponentBase.register_type(AgentBase, 'QAAgent')
class QAAgent(AgentBase):
    conversation: MaxSystemContextConversation = MaxSystemContextConversation(max_size=2)
    type: Literal['QAAgent'] = 'QAAgent'

    def exec(self, 
        input_str: Optional[str] = "",
        llm_kwargs: Optional[Dict] = {} 
        ) -> Any:
        
        self.conversation.add_message(HumanMessage(content=input_str))
        self.llm.predict(conversation=self.conversation, **llm_kwargs)
        
        return self.conversation.get_last().content



import os
import platform
import sys
from datetime import datetime
from importlib.metadata import version, PackageNotFoundError

from IPython import get_ipython
from urllib.parse import unquote


def get_notebook_name():
    """
    Returns the name of the current Jupyter notebook file when running.
    Returns None if not running in a notebook or if the name cannot be determined.
    """
    try:
        ipython = get_ipython()
        if ipython is None or not hasattr(ipython, "kernel"):
            return None

        kernel = ipython.kernel

        if hasattr(kernel, "get_parent"):
            parent = kernel.get_parent()
            metadata = parent.get("metadata", {}) if parent else {}
        else:
            metadata = {}

        def clean_filename(path):
            if not path:
                return None
            filename = os.path.basename(path)
            filename = filename.split("#")[0]
            filename = filename.split("?")[0]
            filename = unquote(filename)
            if filename.endswith(".ipynb"):
                return filename
            return None

        for field in ["filename", "originalPath", "path"]:
            if field in metadata:
                clean_name = clean_filename(metadata[field])
                if clean_name:
                    return clean_name

        cell_id = metadata.get("cellId", "")
        if cell_id and ".ipynb" in cell_id:
            notebook_parts = [part for part in cell_id.split("/") if ".ipynb" in part]
            if notebook_parts:
                clean_name = clean_filename(notebook_parts[0])
                if clean_name:
                    return clean_name
    except Exception as e:
        print(f"Error getting notebook name: {str(e)}")
        return None


def print_notebook_metadata(author_name, github_username):
    print(f"Author: {author_name}")
    print(f"GitHub Username: {github_username}")

    notebook_file = get_notebook_name()

    if notebook_file:
        try:
            last_modified_time = os.path.getmtime(notebook_file)
            last_modified_datetime = datetime.fromtimestamp(last_modified_time)
            print(f"Notebook File: {notebook_file}")
            print(f"Last Modified: {last_modified_datetime}")
        except Exception as e:
            print(f"Could not retrieve last modified datetime: {e}")
    else:
        print("Could not detect the current notebook's filename.")

    print(f"Platform: {platform.system()} {platform.release()}")
    print(f"Python Version: {sys.version}")

    try:
        swarmauri_version = version("swarmauri")
        print(f"Swarmauri Version: {swarmauri_version}")
    except PackageNotFoundError:
        print("Swarmauri version information is unavailable. Ensure it is installed properly.")



from queue import Queue
from typing import Callable, Dict, Any, List
import logging

from swarmauri_base.task_mgmt_strategies.TaskMgmtStrategyBase import TaskMgmtStrategyBase
from swarmauri_core.ComponentBase import ComponentBase

@ComponentBase.register_type(TaskMgmtStrategyBase, 'RoundRobinStrategy')
class RoundRobinStrategy(TaskMgmtStrategyBase):
    """Round-robin task assignment strategy."""

    task_queue: Queue = Queue()  # Synchronous task queue for incoming tasks
    task_assignments: Dict[str, str] = {}  # Tracks task assignments
    current_index: int = 0  # Tracks the next service to assign tasks to

    def assign_task(self, task: Dict[str, Any], service_registry: Callable[[], List[str]]) -> None:
        """
        Assign a task to a service using the round-robin strategy.
        :param task: Task metadata and payload.
        :param service_registry: Callable that returns available services.
        """
        available_services = service_registry()
        if not available_services:
            raise ValueError("No services available for task assignment.")

        # Select the service based on the round-robin index
        service = available_services[self.current_index % len(available_services)]
        self.task_assignments[task["task_id"]] = service
        self.current_index += 1
        logging.info(f"Task '{task['task_id']}' assigned to service '{service}'.")

    def add_task(self, task: Dict[str, Any]) -> None:
        """
        Add a task to the task queue.
        :param task: Task metadata and payload.
        """
        self.task_queue.put(task)

    def remove_task(self, task_id: str) -> None:
        """
        Remove a task from the task registry.
        :param task_id: Unique identifier of the task to remove.
        """
        if task_id in self.task_assignments:
            del self.task_assignments[task_id]
            logging.info(f"Task '{task_id}' removed from assignments.")
        else:
            raise ValueError(f"Task '{task_id}' not found in assignments.")

    def get_task(self, task_id: str) -> Dict[str, Any]:
        """
        Get a task's assigned service.
        :param task_id: Unique identifier of the task.
        :return: Task assignment details.
        """
        if task_id in self.task_assignments:
            service = self.task_assignments[task_id]
            return {"task_id": task_id, "assigned_service": service}
        else:
            raise ValueError(f"Task '{task_id}' not found in assignments.")

    def process_tasks(self, service_registry: Callable[[], List[str]], transport: Callable) -> None:
        """
        Process tasks from the task queue and assign them to services.
        :param service_registry: Callable that returns available services.
        :param transport: Callable used to send tasks to assigned services.
        """
        while not self.task_queue.empty():
            task = self.task_queue.get()
            try:
                self.assign_task(task, service_registry)
                assigned_service = self.task_assignments[task["task_id"]]
                transport.send(task, assigned_service)
            except ValueError as e:
                raise ValueError(f"Error assigning task: {e}")



import inspect

class IndexErrorWithContext(Exception):
    def __init__(self, original_exception):
        self.original_exception = original_exception
        self.stack_info = inspect.stack()
        self.handle_error()

    def handle_error(self):
        # You might want to log this information or handle it differently depending on your application's needs
        frame = self.stack_info[1]  # Assuming the IndexError occurs one level up from where it's caught
        error_details = {
            "message": str(self.original_exception),
            "function": frame.function,
            "file": frame.filename,
            "line": frame.lineno,
            "code_context": ''.join(frame.code_context).strip() if frame.code_context else "No context available"
        }
        print("IndexError occurred with detailed context:")
        for key, value in error_details.items():
            print(f"{key.capitalize()}: {value}")


import asyncio
import json
from typing import AsyncIterator, Iterator, List, Dict, Literal, Optional, Type

import httpx
from pydantic import PrivateAttr
from swarmauri_standard.utils.retry_decorator import retry_on_status_codes
from swarmauri_standard.utils.duration_manager import DurationManager
from swarmauri_standard.conversations.Conversation import Conversation
from swarmauri_standard.messages.AgentMessage import AgentMessage, UsageData
from swarmauri_base.messages.MessageBase import MessageBase
from swarmauri_base.llms.LLMBase import LLMBase
from swarmauri_core.ComponentBase import ComponentBase

@ComponentBase.register_type(LLMBase, 'PerplexityModel')
class PerplexityModel(LLMBase):
    """
    Represents a language model interface for Perplexity API.

    Provides methods for synchronous and asynchronous predictions, streaming,
    and batch processing of conversations using the Perplexity language models.

    Attributes:
        api_key (str): API key for authenticating requests to the Perplexity API.
        allowed_models (List[str]): List of allowed model names that can be used.
        name (str): The default model name to use for predictions.
        type (Literal["PerplexityModel"]): The type identifier for this class.


    Provider resources: https://docs.perplexity.ai/guides/model-cards
    Link to deprecated models: https://docs.perplexity.ai/changelog/changelog#model-deprecation-notice
    """

    api_key: str
    allowed_models: List[str] = [
        "llama-3.1-sonar-small-128k-online",
        "llama-3.1-sonar-large-128k-online",
        "llama-3.1-sonar-huge-128k-online",
        "llama-3.1-sonar-small-128k-chat",
        "llama-3.1-sonar-large-128k-chat",
        "llama-3.1-8b-instruct",
        "llama-3.1-70b-instruct",
    ]
    name: str = "llama-3.1-70b-instruct"
    type: Literal["PerplexityModel"] = "PerplexityModel"
    _client: httpx.Client = PrivateAttr(default=None)
    _async_client: httpx.AsyncClient = PrivateAttr(default=None)
    _BASE_URL: str = PrivateAttr(default="https://api.perplexity.ai/chat/completions")

    def __init__(self, **data):
        """
        Initialize the GroqAIAudio class with the provided data.

        Args:
            **data: Arbitrary keyword arguments containing initialization data.
        """
        super().__init__(**data)
        self._client = httpx.Client(
            headers={"Authorization": f"Bearer {self.api_key}"},
            base_url=self._BASE_URL,
            timeout=30,
        )
        self._async_client = httpx.AsyncClient(
            headers={"Authorization": f"Bearer {self.api_key}"},
            base_url=self._BASE_URL,
            timeout=30,
        )

    def _format_messages(
        self, messages: List[Type[MessageBase]]
    ) -> List[Dict[str, str]]:
        """
        Formats the list of message objects for the API request.

        Args:
            messages: A list of message objects.

        Returns:
            A list of formatted message dictionaries.
        """
        message_properties = ["content", "role", "name"]
        formatted_messages = [
            message.model_dump(include=message_properties, exclude_none=True)
            for message in messages
        ]
        return formatted_messages

    def _prepare_usage_data(
        self,
        usage_data,
        prompt_time: float = 0,
        completion_time: float = 0,
    ) -> UsageData:
        """
        Prepares usage data and calculates response timing.

        Args:
            usage_data: The raw usage data from the API response.
            prompt_time: Time taken for the prompt processing.
            completion_time: Time taken for the completion processing.

        Returns:
            A UsageData object containing token and timing information.
        """

        total_time = prompt_time + completion_time

        usage = UsageData(
            prompt_tokens=usage_data.get("prompt_tokens", 0),
            completion_tokens=usage_data.get("completion_tokens", 0),
            total_tokens=usage_data.get("total_tokens", 0),
            prompt_time=prompt_time,
            completion_time=completion_time,
            total_time=total_time,
        )

        return usage

    @retry_on_status_codes((429, 529), max_retries=1)
    def predict(
        self,
        conversation: Conversation,
        temperature=0.7,
        max_tokens=256,
        top_p: Optional[float] = None,
        top_k: Optional[int] = None,
        return_citations: Optional[bool] = False,
        presence_penalty: Optional[float] = None,
        frequency_penalty: Optional[float] = None,
    ) -> Conversation:
        """
        Makes a synchronous prediction request.

        Args:
            conversation: The conversation object containing the history.
            temperature: Sampling temperature for response generation.
            max_tokens: Maximum number of tokens for the response.
            top_p: Nucleus sampling parameter.
            top_k: Top-k sampling parameter.
            return_citations: Whether to return citations in the response.
            presence_penalty: Penalty for new tokens based on presence.
            frequency_penalty: Penalty for new tokens based on frequency.

        Returns:
            An updated Conversation object with the model's response.
        """

        if top_p and top_k:
            raise ValueError("Do not set top_p and top_k")

        formatted_messages = self._format_messages(conversation.history)

        payload = {
            "model": self.name,
            "messages": formatted_messages,
            "max_tokens": max_tokens,
            "temperature": temperature,
            "top_p": top_p,
            "return_citations": return_citations,
            "top_k": top_k,
            "presence_penalty": presence_penalty,
            "frequency_penalty": frequency_penalty,
        }
        headers = {
            "accept": "application/json",
            "content-type": "application/json",
            "authorization": f"Bearer {self.api_key}",
        }

        with DurationManager() as prompt_timer:
            response = self._client.post(self._BASE_URL, json=payload, headers=headers)
            response.raise_for_status()

        result = response.json()
        message_content = result["choices"][0]["message"]["content"]

        usage_data = result.get("usage", {})

        usage = self._prepare_usage_data(usage_data, prompt_timer.duration)

        conversation.add_message(AgentMessage(content=message_content, usage=usage))
        return conversation

    @retry_on_status_codes((429, 529), max_retries=1)
    async def apredict(
        self,
        conversation: Conversation,
        temperature=0.7,
        max_tokens=256,
        top_p: Optional[float] = None,
        top_k: Optional[int] = None,
        return_citations: Optional[bool] = False,
        presence_penalty: Optional[float] = None,
        frequency_penalty: Optional[float] = None,
    ) -> Conversation:
        """
        Makes an asynchronous prediction request.

        Args:
            conversation: The conversation object containing the history.
            temperature: Sampling temperature for response generation.
            max_tokens: Maximum number of tokens for the response.
            top_p: Nucleus sampling parameter.
            top_k: Top-k sampling parameter.
            return_citations: Whether to return citations in the response.
            presence_penalty: Penalty for new tokens based on presence.
            frequency_penalty: Penalty for new tokens based on frequency.

        Returns:
            An updated Conversation object with the model's response.
        """

        if top_p and top_k:
            raise ValueError("Do not set top_p and top_k")

        formatted_messages = self._format_messages(conversation.history)

        payload = {
            "model": self.name,
            "messages": formatted_messages,
            "max_tokens": max_tokens,
            "temperature": temperature,
            "top_p": top_p,
            "return_citations": return_citations,
            "top_k": top_k,
            "presence_penalty": presence_penalty,
            "frequency_penalty": frequency_penalty,
        }
        headers = {
            "accept": "application/json",
            "content-type": "application/json",
            "authorization": f"Bearer {self.api_key}",
        }

        with DurationManager() as prompt_timer:
            response = await self._async_client.post(
                self._BASE_URL, json=payload, headers=headers
            )
            response.raise_for_status()

        result = response.json()

        message_content = result["choices"][0]["message"]["content"]

        usage_data = result.get("usage", {})
        usage = self._prepare_usage_data(usage_data, prompt_timer.duration)
        conversation.add_message(AgentMessage(content=message_content, usage=usage))

        return conversation

    @retry_on_status_codes((429, 529), max_retries=1)
    def stream(
        self,
        conversation: Conversation,
        temperature=0.7,
        max_tokens=256,
        top_p: Optional[float] = None,
        top_k: Optional[int] = None,
        return_citations: Optional[bool] = False,
        presence_penalty: Optional[float] = None,
        frequency_penalty: Optional[float] = None,
    ) -> Iterator[str]:
        """
        Synchronously streams the response for a given conversation.

        Args:
            conversation (Conversation): The conversation object containing message history.
            temperature (float, optional): Sampling temperature for response generation. Defaults to 0.7.
            max_tokens (int, optional): Maximum number of tokens in the generated response. Defaults to 256.
            top_p (Optional[float], optional): Nucleus sampling parameter. If specified, `top_k` should not be set.
            top_k (Optional[int], optional): Top-k sampling parameter. If specified, `top_p` should not be set.
            return_citations (Optional[bool], optional): Whether to return citations in the response. Defaults to False.
            presence_penalty (Optional[float], optional): Penalty for introducing new topics. Defaults to None.
            frequency_penalty (Optional[float], optional): Penalty for repeating existing tokens. Defaults to None.

        Yields:
            str: Chunks of response content as the data is streamed.
        """
        if top_p and top_k:
            raise ValueError("Do not set top_p and top_k")

        formatted_messages = self._format_messages(conversation.history)

        url = "https://api.perplexity.ai/chat/completions"

        payload = {
            "model": self.name,
            "messages": formatted_messages,
            "max_tokens": max_tokens,
            "temperature": temperature,
            "top_p": top_p,
            "return_citations": return_citations,
            "top_k": top_k,
            "presence_penalty": presence_penalty,
            "frequency_penalty": frequency_penalty,
            "stream": True,
        }
        headers = {
            "accept": "application/json",
            "content-type": "application/json",
            "authorization": f"Bearer {self.api_key}",
        }

        with DurationManager() as prompt_timer:
            response = self._client.post(url, json=payload, headers=headers)
            response.raise_for_status()

        message_content = ""

        with DurationManager() as completion_timer:
            for chunk in response.iter_lines():
                json_string = chunk.replace("data: ", "", 1)
                if json_string:
                    chunk_data = json.loads(json_string)
                    delta_content = (
                        chunk_data.get("choices", [{}])[0]
                        .get("delta", {})
                        .get("content", "")
                    )
                    message_content += delta_content
                    yield delta_content
                    if chunk_data["usage"]:
                        usage_data = chunk_data["usage"]

        usage = self._prepare_usage_data(
            usage_data, prompt_timer.duration, completion_timer.duration
        )

        conversation.add_message(AgentMessage(content=message_content, usage=usage))

    @retry_on_status_codes((429, 529), max_retries=1)
    async def astream(
        self,
        conversation: Conversation,
        temperature=0.7,
        max_tokens=256,
        top_p: Optional[float] = None,
        top_k: Optional[int] = None,
        return_citations: Optional[bool] = False,
        presence_penalty: Optional[float] = None,
        frequency_penalty: Optional[float] = None,
    ) -> AsyncIterator[str]:
        """
        Asynchronously streams the response for a given conversation.

        Args:
            conversation (Conversation): The conversation object containing message history.
            temperature (float, optional): Sampling temperature for response generation. Defaults to 0.7.
            max_tokens (int, optional): Maximum number of tokens in the generated response. Defaults to 256.
            top_p (Optional[float], optional): Nucleus sampling parameter. If specified, `top_k` should not be set.
            top_k (Optional[int], optional): Top-k sampling parameter. If specified, `top_p` should not be set.
            return_citations (Optional[bool], optional): Whether to return citations in the response. Defaults to False.
            presence_penalty (Optional[float], optional): Penalty for introducing new topics. Defaults to None.
            frequency_penalty (Optional[float], optional): Penalty for repeating existing tokens. Defaults to None.

        Yields:
            str: Chunks of response content as the data is streamed asynchronously.
        """
        if top_p and top_k:
            raise ValueError("Do not set top_p and top_k")

        formatted_messages = self._format_messages(conversation.history)

        payload = {
            "model": self.name,
            "messages": formatted_messages,
            "max_tokens": max_tokens,
            "temperature": temperature,
            "top_p": top_p,
            "return_citations": return_citations,
            "top_k": top_k,
            "presence_penalty": presence_penalty,
            "frequency_penalty": frequency_penalty,
            "stream": True,
        }

        with DurationManager() as prompt_timer:
            response = await self._async_client.post(self._BASE_URL, json=payload)
            response.raise_for_status()

        message_content = ""
        usage_data = {}

        with DurationManager() as completion_timer:
            async for line in response.aiter_lines():
                json_string = line.replace("data: ", "", 1)
                if json_string:  # Ensure it's not empty
                    chunk_data = json.loads(json_string)
                    delta_content = (
                        chunk_data.get("choices", [{}])[0]
                        .get("delta", {})
                        .get("content", "")
                    )
                    message_content += delta_content
                    yield delta_content
                    usage_data = chunk_data.get("usage", usage_data)

        usage = self._prepare_usage_data(
            usage_data, prompt_timer.duration, completion_timer.duration
        )
        conversation.add_message(AgentMessage(content=message_content, usage=usage))

    def batch(
        self,
        conversations: List[Conversation],
        temperature=0.7,
        max_tokens=256,
        top_p: Optional[float] = None,
        top_k: Optional[int] = None,
        return_citations: Optional[bool] = False,
        presence_penalty: Optional[float] = None,
        frequency_penalty: Optional[float] = None,
    ) -> List[Conversation]:
        """
        Processes a batch of conversations synchronously.

        Args:
            conversations (List[Conversation]): List of conversation objects.
            temperature (float, optional): Sampling temperature for response generation. Defaults to 0.7.
            max_tokens (int, optional): Maximum number of tokens in the generated response. Defaults to 256.
            top_p (Optional[float], optional): Nucleus sampling parameter. If specified, `top_k` should not be set.
            top_k (Optional[int], optional): Top-k sampling parameter. If specified, `top_p` should not be set.
            return_citations (Optional[bool], optional): Whether to return citations in the response. Defaults to False.
            presence_penalty (Optional[float], optional): Penalty for introducing new topics. Defaults to None.
            frequency_penalty (Optional[float], optional): Penalty for repeating existing tokens. Defaults to None.

        Returns:
            List[Conversation]: List of updated conversation objects after processing.
        """
        return [
            self.predict(
                conversation=conv,
                temperature=temperature,
                max_tokens=max_tokens,
                top_p=top_p,
                top_k=top_k,
                return_citations=return_citations,
                presence_penalty=presence_penalty,
                frequency_penalty=frequency_penalty,
            )
            for conv in conversations
        ]

    async def abatch(
        self,
        conversations: List[Conversation],
        temperature=0.7,
        max_tokens=256,
        top_p: Optional[float] = None,
        top_k: Optional[int] = None,
        return_citations: Optional[bool] = False,
        presence_penalty: Optional[float] = None,
        frequency_penalty: Optional[float] = None,
        max_concurrent: int = 5,  # Maximum concurrent tasks
    ) -> List[Conversation]:
        """
        Asynchronously processes a batch of conversations with a limit on concurrent tasks.

        Args:
            conversations (List[Conversation]): List of conversation objects.
            temperature (float, optional): Sampling temperature for response generation. Defaults to 0.7.
            max_tokens (int, optional): Maximum number of tokens in the generated response. Defaults to 256.
            top_p (Optional[float], optional): Nucleus sampling parameter. If specified, `top_k` should not be set.
            top_k (Optional[int], optional): Top-k sampling parameter. If specified, `top_p` should not be set.
            return_citations (Optional[bool], optional): Whether to return citations in the response. Defaults to False.
            presence_penalty (Optional[float], optional): Penalty for introducing new topics. Defaults to None.
            frequency_penalty (Optional[float], optional): Penalty for repeating existing tokens. Defaults to None.
            max_concurrent (int, optional): Maximum number of concurrent tasks. Defaults to 5.

        Returns:
            List[Conversation]: List of updated conversation objects after processing asynchronously.
        """
        semaphore = asyncio.Semaphore(max_concurrent)

        async def process_conversation(conv) -> Conversation:
            async with semaphore:
                return await self.apredict(
                    conversation=conv,
                    temperature=temperature,
                    max_tokens=max_tokens,
                    top_p=top_p,
                    top_k=top_k,
                    return_citations=return_citations,
                    presence_penalty=presence_penalty,
                    frequency_penalty=frequency_penalty,
                )

        tasks = [process_conversation(conv) for conv in conversations]
        return await asyncio.gather(*tasks)



# from swarmauri_standard.utils._lazy_import import _lazy_import

# # List of agent names (file names without the ".py" extension) and corresponding class names
# agent_files = [
#     ("swarmauri_standard.agents.QAAgent", "QAAgent"),
#     ("swarmauri_standard.agents.RagAgent", "RagAgent"),
#     ("swarmauri_standard.agents.SimpleConversationAgent", "SimpleConversationAgent"),
#     ("swarmauri_standard.agents.ToolAgent", "ToolAgent"),
# ]

# # Lazy loading of agent classes, storing them in variables
# for module_name, class_name in agent_files:
#     globals()[class_name] = _lazy_import(module_name, class_name)

# # Adding the lazy-loaded agent classes to __all__
# __all__ = [class_name for _, class_name in agent_files]



# from swarmauri_standard.utils._lazy_import import _lazy_import

# # List of distances names (file names without the ".py" extension) and corresponding class names
# distances_files = [
#     ("swarmauri_standard.distances.CanberraDistance", "CanberraDistance"),
#     ("swarmauri_standard.distances.ChebyshevDistance", "ChebyshevDistance"),
#     ("swarmauri_standard.distances.ChiSquaredDistance", "ChiSquaredDistance"),
#     ("swarmauri_standard.distances.CosineDistance", "CosineDistance"),
#     ("swarmauri_standard.distances.EuclideanDistance", "EuclideanDistance"),
#     ("swarmauri_standard.distances.HaversineDistance", "HaversineDistance"),
#     ("swarmauri_standard.distances.JaccardIndexDistance", "JaccardIndexDistance"),
#     ("swarmauri_standard.distances.LevenshteinDistance", "LevenshteinDistance"),
#     ("swarmauri_standard.distances.ManhattanDistance", "ManhattanDistance"),
#     ("swarmauri_standard.distances.MinkowskiDistance", "MinkowskiDistance"),
#     ("swarmauri_standard.distances.SorensenDiceDistance", "SorensenDiceDistance"),
#     (
#         "swarmauri_standard.distances.SquaredEuclideanDistance",
#         "SquaredEuclideanDistance",
#     ),
# ]

# # Lazy loading of distances classes, storing them in variables
# for module_name, class_name in distances_files:
#     globals()[class_name] = _lazy_import(module_name, class_name)

# # Adding the lazy-loaded distances classes to __all__
# __all__ = [class_name for _, class_name in distances_files]



import re
from typing import Any, Dict, List, Literal
from swarmauri_standard.tools.Parameter import Parameter
from swarmauri_base.tools.ToolBase import ToolBase
from swarmauri_core.ComponentBase import ComponentBase

@ComponentBase.register_type(ToolBase, 'ColemanLiauIndexTool')
class ColemanLiauIndexTool(ToolBase):
    """
    A tool for calculating the Coleman-Liau Index (CLI).

    Attributes:
        version (str): The version of the tool.
        name (str): The name of the tool.
        type (Literal["ColemanLiauIndexTool"]): The type of the tool.
        description (str): A brief description of what the tool does.
        parameters (List[Parameter]): The parameters for configuring the tool.
    """

    version: str = "0.1.dev1"
    name: str = "ColemanLiauIndexTool"
    type: Literal["ColemanLiauIndexTool"] = "ColemanLiauIndexTool"
    description: str = "Calculates the Coleman-Liau Index (CLI) for a given text."
    parameters: List[Parameter] = [
        Parameter(
            name="input_text",
            type="string",
            description="The input text for which to calculate the CLI.",
            required=True,
        )
    ]

    def __call__(self, data: Dict[str, Any]) -> Dict[str, float]:
        """
        Executes the CLI tool and returns the readability score.

        Coleman-Liau Index formula:
        CLI = 0.0588 * (characters/words * 100) - 0.296 * (sentences/words * 100) - 15.8

        Parameters:
            data (Dict[str, Any]): The input data containing "input_text".

        Returns:
            float: The Coleman-Liau Index.

        Raises:
            ValueError: If the input data is invalid.
        """
        if self.validate_input(data):
            text = data["input_text"]
            num_sentences = self.count_sentences(text)
            num_words = self.count_words(text)
            num_characters = self.count_characters(text)
            if num_sentences == 0 or num_words == 0:
                return {"coleman_liau_index": 0.0}
            L = (
                num_characters / num_words
            ) * 100  # Average number of letters per 100 words
            S = (
                num_sentences / num_words
            ) * 100  # Average number of sentences per 100 words
            cli_score = 0.0588 * L - 0.296 * S - 15.8
            return {"coleman_liau_index": cli_score}
        else:
            raise ValueError("Invalid input for ColemanLiauIndexTool.")

    def count_sentences(self, text: str) -> int:
        """
        Counts the number of sentences in the text.

        Parameters:
            text (str): The input text.

        Returns:
            int: The number of sentences in the text.
        """
        sentence_endings = re.compile(r"[.!?]")
        sentences = sentence_endings.split(text)
        return len([s for s in sentences if s.strip()])  # Count non-empty sentences

    def count_words(self, text: str) -> int:
        """
        Counts the number of words in the text.

        Parameters:
            text (str): The input text.

        Returns:
            int: The number of words in the text.
        """
        words = re.findall(r"\b\w+\b", text)
        return len(words)

    def count_characters(self, text: str) -> int:
        """
        Counts the number of characters in the text.

        Parameters:
            text (str): The input text.

        Returns:
            int: The number of characters in the text.
        """
        return len(re.findall(r"[A-Za-z]", text))  # Count characters excluding spaces

    def validate_input(self, data: Dict[str, Any]) -> bool:
        """
        Validates the input data to ensure it contains the necessary 'input_text' key.

        Parameters:
            data (Dict[str, Any]): The input data to validate.

        Returns:
            bool: True if the input is valid, False otherwise.
        """
        return (
            isinstance(data, dict)
            and "input_text" in data
            and isinstance(data["input_text"], str)
        )



from typing import Literal
from swarmauri_base.documents.DocumentBase import DocumentBase
from swarmauri_core.ComponentBase import ComponentBase

@ComponentBase.register_type(DocumentBase, 'Document')
class Document(DocumentBase):
    type: Literal['Document'] = 'Document'



import json
import httpx
import asyncio
from typing import List, Dict, Literal, AsyncIterator, Iterator, Type
from pydantic import PrivateAttr

from swarmauri_standard.utils.retry_decorator import retry_on_status_codes
from swarmauri_standard.messages.AgentMessage import AgentMessage, UsageData
from swarmauri_base.messages.MessageBase import MessageBase
from swarmauri_base.llms.LLMBase import LLMBase
from swarmauri_core.ComponentBase import ComponentBase

@ComponentBase.register_type(LLMBase, 'DeepInfraModel')
class DeepInfraModel(LLMBase):
    """
    A class for interacting with DeepInfra's model API for text generation.

    This implementation uses httpx for both synchronous and asynchronous HTTP requests,
    providing support for predictions, streaming responses, and batch processing.

    Attributes:
        api_key (str): DeepInfra API key for authentication
            Can be obtained from: https://deepinfra.com/dash/api_keys

        allowed_models (List[str]): List of supported model identifiers on DeepInfra
            Full list available at: https://deepinfra.com/models/text-generation

        name (str): The currently selected model name
            Defaults to "Qwen/Qwen2-72B-Instruct"

        type (Literal["DeepInfraModel"]): Type identifier for the model class

    Link to Allowed Models: https://deepinfra.com/models/text-generation
    Link to API KEY: https://deepinfra.com/dash/api_keys
    """

    _BASE_URL: str = PrivateAttr("https://api.deepinfra.com/v1/openai")
    _client: httpx.Client = PrivateAttr(default=None)
    _async_client: httpx.AsyncClient = PrivateAttr(default=None)

    api_key: str
    allowed_models: List[str] = [
        "01-ai/Yi-34B-Chat",
        "Gryphe/MythoMax-L2-13b",  # not consistent with results
        "HuggingFaceH4/zephyr-orpo-141b-A35b-v0.1",
        "Phind/Phind-CodeLlama-34B-v2",
        "Qwen/Qwen2-72B-Instruct",
        "Qwen/Qwen2-7B-Instruct",
        "Qwen/Qwen2.5-72B-Instruct",
        "Sao10K/L3-70B-Euryale-v2.1",
        "Sao10K/L3.1-70B-Euryale-v2.2",
        "bigcode/starcoder2-15b",
        "bigcode/starcoder2-15b-instruct-v0.1",
        "codellama/CodeLlama-34b-Instruct-hf",
        "codellama/CodeLlama-70b-Instruct-hf",
        "cognitivecomputations/dolphin-2.6-mixtral-8x7b",
        "cognitivecomputations/dolphin-2.9.1-llama-3-70b",
        "databricks/dbrx-instruct",
        "google/codegemma-7b-it",
        "google/gemma-1.1-7b-it",
        "google/gemma-2-27b-it",
        "google/gemma-2-9b-it",
        "lizpreciatior/lzlv_70b_fp16_hf",  # not consistent with results
        "mattshumer/Reflection-Llama-3.1-70B",
        "mattshumer/Reflection-Llama-3.1-70B",
        "meta-llama/Llama-2-13b-chat-hf",
        "meta-llama/Llama-2-70b-chat-hf",
        "meta-llama/Llama-2-7b-chat-hf",
        "meta-llama/Meta-Llama-3-70B-Instruct",
        "meta-llama/Meta-Llama-3-8B-Instruct",
        "meta-llama/Meta-Llama-3.1-405B-Instruct",
        "meta-llama/Meta-Llama-3.1-70B-Instruct",
        "meta-llama/Meta-Llama-3.1-8B-Instruct",
        "microsoft/Phi-3-medium-4k-instruct",
        "microsoft/WizardLM-2-7B",
        "microsoft/WizardLM-2-8x22B",
        "mistralai/Mistral-7B-Instruct-v0.1",
        "mistralai/Mistral-7B-Instruct-v0.2",
        "mistralai/Mistral-7B-Instruct-v0.3",
        "mistralai/Mistral-Nemo-Instruct-2407",
        "mistralai/Mixtral-8x22B-Instruct-v0.1",
        "mistralai/Mixtral-8x22B-v0.1",
        "mistralai/Mixtral-8x22B-v0.1",
        "mistralai/Mixtral-8x7B-Instruct-v0.1",
        "nvidia/Nemotron-4-340B-Instruct",
        "openbmb/MiniCPM-Llama3-V-2_5",
        "openchat/openchat-3.6-8b",
        "openchat/openchat_3.5",  # not compliant with system context
        # "deepinfra/airoboros-70b", # deprecated: https://deepinfra.com/deepinfra/airoboros-70b
        #  'Gryphe/MythoMax-L2-13b-turbo', # deprecated: https://deepinfra.com/Gryphe/MythoMax-L2-13b-turbo/api
        # "Austism/chronos-hermes-13b-v2",  # deprecating: https://deepinfra.com/Austism/chronos-hermes-13b-v2/api
    ]

    name: str = "Qwen/Qwen2-72B-Instruct"
    type: Literal["DeepInfraModel"] = "DeepInfraModel"

    def __init__(self, **data):
        """
        Initializes the DeepInfraModel instance with the provided API key
        and sets up httpx clients for both sync and async operations.

        Args:
            **data: Keyword arguments for model initialization.
        """
        super().__init__(**data)
        headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self.api_key}",
        }
        self._client = httpx.Client(
            headers=headers, base_url=self._BASE_URL, timeout=30
        )
        self._async_client = httpx.AsyncClient(
            headers=headers, base_url=self._BASE_URL, timeout=30
        )

    def _format_messages(
        self, messages: List[SubclassUnion[MessageBase]]
    ) -> List[Dict[str, str]]:
        """
        Formats conversation history into a list of dictionaries suitable for API requests.

        Args:
            messages (List[SubclassUnion[MessageBase]]): The conversation history.

        Returns:
            List[Dict[str, str]]: Formatted message list.
        """
        message_properties = ["content", "role", "name"]
        return [
            message.model_dump(include=message_properties, exclude_none=True)
            for message in messages
        ]

    def _create_request_payload(
        self,
        formatted_messages: List[Dict[str, str]],
        temperature: float,
        max_tokens: int,
        enable_json: bool,
        stop: List[str] = None,
        stream: bool = False,
    ) -> Dict:
        """
        Creates the payload for the API request.

        Args:
            formatted_messages (List[Dict[str, str]]): Formatted messages for the conversation.
            temperature (float): Sampling temperature for the response.
            max_tokens (int): Maximum number of tokens to generate.
            enable_json (bool): Whether to enable JSON response format.
            stop (List[str], optional): Stop sequences.
            stream (bool): Whether to stream the response.

        Returns:
            Dict: Payload for the API request.
        """
        payload = {
            "model": self.name,
            "messages": formatted_messages,
            "temperature": temperature,
            "max_tokens": max_tokens,
            "top_p": 1,
            "frequency_penalty": 0,
            "presence_penalty": 0,
            "stream": stream,
        }

        if stop:
            payload["stop"] = stop

        if enable_json:
            payload["response_format"] = {"type": "json_object"}

        return payload

    @retry_on_status_codes((429, 529), max_retries=1)
    def predict(
        self,
        conversation,
        temperature=0.7,
        max_tokens=256,
        enable_json=False,
        stop: List[str] = None,
    ):
        """
        Sends a synchronous request to generate a response from the model.

        Args:
            conversation: The conversation object containing message history.
            temperature (float): Sampling temperature for response generation.
            max_tokens (int): Maximum number of tokens to generate.
            enable_json (bool): Flag for enabling JSON response format.
            stop (List[str], optional): Stop sequences for the response.

        Returns:
            Updated conversation with the model's response.
        """
        formatted_messages = self._format_messages(conversation.history)
        payload = self._create_request_payload(
            formatted_messages, temperature, max_tokens, enable_json, stop
        )

        response = self._client.post("/chat/completions", json=payload)
        response.raise_for_status()

        result = response.json()
        message_content = result["choices"][0]["message"]["content"]
        conversation.add_message(AgentMessage(content=message_content))

        return conversation

    @retry_on_status_codes((429, 529), max_retries=1)
    async def apredict(
        self,
        conversation,
        temperature=0.7,
        max_tokens=256,
        enable_json=False,
        stop: List[str] = None,
    ):
        """
        Sends an asynchronous request to generate a response from the model.

        Args:
            conversation: The conversation object containing message history.
            temperature (float): Sampling temperature for response generation.
            max_tokens (int): Maximum number of tokens to generate.
            enable_json (bool): Flag for enabling JSON response format.
            stop (List[str], optional): Stop sequences for the response.

        Returns:
            Updated conversation with the model's response.
        """
        formatted_messages = self._format_messages(conversation.history)
        payload = self._create_request_payload(
            formatted_messages, temperature, max_tokens, enable_json, stop
        )

        response = await self._async_client.post("/chat/completions", json=payload)
        response.raise_for_status()

        result = response.json()
        message_content = result["choices"][0]["message"]["content"]
        conversation.add_message(AgentMessage(content=message_content))

        return conversation

    @retry_on_status_codes((429, 529), max_retries=1)
    def stream(
        self,
        conversation,
        temperature=0.7,
        max_tokens=256,
        stop: List[str] = None,
    ) -> Iterator[str]:
        """
        Streams response content from the model synchronously.

        Args:
            conversation: The conversation object containing message history.
            temperature (float): Sampling temperature for response generation.
            max_tokens (int): Maximum number of tokens to generate.
            stop (List[str], optional): Stop sequences for the response.

        Yields:
            str: Chunks of content from the model's response.
        """
        formatted_messages = self._format_messages(conversation.history)
        payload = self._create_request_payload(
            formatted_messages, temperature, max_tokens, False, stop, stream=True
        )

        with self._client.stream("POST", "/chat/completions", json=payload) as response:
            response.raise_for_status()
            collected_content = []

            for line in response.iter_lines():
                # Convert bytes to string if necessary
                if isinstance(line, bytes):
                    line = line.decode("utf-8")

                if line.startswith("data: "):
                    line = line[6:]  # Remove 'data: ' prefix
                    if line != "[DONE]":
                        chunk = json.loads(line)
                        if chunk["choices"][0]["delta"].get("content"):
                            content = chunk["choices"][0]["delta"]["content"]
                            collected_content.append(content)
                            yield content

        full_content = "".join(collected_content)
        conversation.add_message(AgentMessage(content=full_content))

    @retry_on_status_codes((429, 529), max_retries=1)
    async def astream(
        self,
        conversation,
        temperature=0.7,
        max_tokens=256,
        stop: List[str] = None,
    ) -> AsyncIterator[str]:
        """
        Streams response content from the model asynchronously.

        Args:
            conversation: The conversation object containing message history.
            temperature (float): Sampling temperature for response generation.
            max_tokens (int): Maximum number of tokens to generate.
            stop (List[str], optional): Stop sequences for the response.

        Yields:
            str: Chunks of content from the model's response.
        """
        formatted_messages = self._format_messages(conversation.history)
        payload = self._create_request_payload(
            formatted_messages, temperature, max_tokens, False, stop, stream=True
        )

        async with self._async_client.stream(
            "POST", "/chat/completions", json=payload
        ) as response:
            response.raise_for_status()
            collected_content = []

            async for line in response.aiter_lines():
                if line.startswith("data: "):
                    line = line[6:]  # Remove 'data: ' prefix
                    if line != "[DONE]":
                        chunk = json.loads(line)
                        if chunk["choices"][0]["delta"].get("content"):
                            content = chunk["choices"][0]["delta"]["content"]
                            collected_content.append(content)
                            yield content

        full_content = "".join(collected_content)
        conversation.add_message(AgentMessage(content=full_content))

    def batch(
        self,
        conversations: List,
        temperature=0.7,
        max_tokens=256,
        enable_json=False,
        stop: List[str] = None,
    ) -> List:
        """
        Processes multiple conversations in batch synchronously.

        Args:
            conversations (List): List of conversation objects.
            temperature (float): Sampling temperature for response generation.
            max_tokens (int): Maximum number of tokens to generate.
            enable_json (bool): Flag for enabling JSON response format.
            stop (List[str], optional): Stop sequences for responses.

        Returns:
            List: List of updated conversations with model responses.
        """
        return [
            self.predict(
                conv,
                temperature=temperature,
                max_tokens=max_tokens,
                enable_json=enable_json,
                stop=stop,
            )
            for conv in conversations
        ]

    async def abatch(
        self,
        conversations: List,
        temperature=0.7,
        max_tokens=256,
        enable_json=False,
        stop: List[str] = None,
        max_concurrent=5,
    ) -> List:
        """
        Processes multiple conversations asynchronously, with concurrency control.

        Args:
            conversations (List): List of conversation objects.
            temperature (float): Sampling temperature for response generation.
            max_tokens (int): Maximum number of tokens to generate.
            enable_json (bool): Flag for enabling JSON response format.
            stop (List[str], optional): Stop sequences for responses.
            max_concurrent (int): Maximum number of concurrent tasks.

        Returns:
            List: List of updated conversations with model responses.
        """
        semaphore = asyncio.Semaphore(max_concurrent)

        async def process_conversation(conv):
            async with semaphore:
                return await self.apredict(
                    conv,
                    temperature=temperature,
                    max_tokens=max_tokens,
                    enable_json=enable_json,
                    stop=stop,
                )

        tasks = [process_conversation(conv) for conv in conversations]
        return await asyncio.gather(*tasks)



from typing import Literal
from swarmauri_base.prompts.PromptMatrixBase import PromptMatrixBase
from swarmauri_core.ComponentBase import ComponentBase

@ComponentBase.register_type(PromptMatrixBase, 'PromptMatrix')
class PromptMatrix(PromptMatrixBase):
    type: Literal['PromptMatrix'] = 'PromptMatrix'


from typing import Literal, Optional, Any
from pydantic import Field
from swarmauri_base.messages.MessageBase import MessageBase
from swarmauri_core.ComponentBase import ComponentBase

@ComponentBase.register_type(MessageBase, 'ZeroMeasurement')
class FunctionMessage(MessageBase):
    content: str
    role: str = Field(default="tool")
    tool_call_id: str
    name: str
    type: Literal["FunctionMessage"] = "FunctionMessage"
    usage: Optional[Any] = (
        None  # 🚧 Placeholder for CompletionUsage(input_tokens, output_tokens, completion time, etc)
    )




import base64
import matplotlib.pyplot as plt
from typing import List, Literal
from pydantic import Field
from swarmauri_base.tools.ToolBase import ToolBase
from swarmauri_standard.tools.Parameter import Parameter
from swarmauri_base.tools.ToolBase import ToolBase
from swarmauri_core.ComponentBase import ComponentBase

@ComponentBase.register_type(ToolBase, 'MatplotlibTool')
class MatplotlibTool(ToolBase):
    version: str = "1.0.0"
    name: str = "MatplotlibTool"
    description: str = (
        "Generates a plot using Matplotlib library based on provided configuration."
    )
    type: Literal["MatplotlibTool"] = "MatplotlibTool"

    parameters: List[Parameter] = Field(
        default_factory=lambda: [
            Parameter(
                name="plot_type",
                type="string",
                description="Type of plot to generate (e.g., 'line', 'bar', 'scatter').",
                required=True,
                enum=["line", "bar", "scatter"],
            ),
            Parameter(
                name="x_data",
                type="list<float>",
                description="X-axis data for the plot.",
                required=True,
            ),
            Parameter(
                name="y_data",
                type="list<float>",
                description="Y-axis data for the plot.",
                required=True,
            ),
            Parameter(
                name="title",
                type="string",
                description="Title of the plot.",
                required=False,
                default="",
            ),
            Parameter(
                name="x_label",
                type="string",
                description="Label for the X-axis.",
                required=False,
                default="",
            ),
            Parameter(
                name="y_label",
                type="string",
                description="Label for the Y-axis.",
                required=False,
                default="",
            ),
            Parameter(
                name="save_path",
                type="string",
                description="Path to save the generated plot image.",
                required=False,
                default="plot.png",
            ),
        ]
    )

    def __call__(
        self,
        plot_type: str,
        x_data: List[float],
        y_data: List[float],
        title: str = "",
        x_label: str = "",
        y_label: str = "",
        save_path: str = "plot.png",
    ):
        """
        Generates a plot using Matplotlib based on provided configuration.

        Parameters:
            plot_type (str): The type of the plot ('line', 'bar', 'scatter').
            x_data (List[float]): X-axis data for the plot.
            y_data (List[float]): Y-axis data for the plot.
            title (str): Title of the plot.
            x_label (str): Label for the X-axis.
            y_label (str): Label for the Y-axis.
            save_path (str): Path to save the generated plot image.

        Returns:
            str: Path where the plot image is saved.
        """
        plt.figure()

        if plot_type == "line":
            plt.plot(x_data, y_data)
        elif plot_type == "bar":
            plt.bar(x_data, y_data)
        elif plot_type == "scatter":
            plt.scatter(x_data, y_data)
        else:
            raise ValueError(f"Unsupported plot type: {plot_type}")

        if title:
            plt.title(title)
        if x_label:
            plt.xlabel(x_label)
        if y_label:
            plt.ylabel(y_label)

        plt.savefig(save_path)
        plt.close()

        with open(save_path, "rb") as image_file:
            encoded_image = base64.b64encode(image_file.read()).decode("utf-8")

        return {"img_path": save_path, "img_base64": encoded_image, "data": []}




from copy import deepcopy
from typing import Dict, Any
from pydantic import Field, model_validator
from swarmauri_base.state.StateBase import StateBase
from swarmauri_core.ComponentBase import ComponentBase

@ComponentBase.register_type(StateBase, 'DictState')
class DictState(StateBase):
    """
    A concrete implementation of StateBase that manages state as a dictionary.
    """

    state_data: Dict[str, Any] = Field(
        default_factory=dict, description="The current state data."
    )

    def read(self) -> Dict[str, Any]:
        """
        Reads and returns the current state as a dictionary.
        """
        return deepcopy(self.state_data)

    def write(self, data: Dict[str, Any]) -> None:
        """
        Replaces the current state with the given data.
        """
        self.state_data = deepcopy(data)

    def update(self, data: Dict[str, Any]) -> None:
        """
        Updates the state with the given data.
        """
        self.state_data.update(data)

    def reset(self) -> None:
        """
        Resets the state to an empty dictionary.
        """
        self.state_data = {}

    def deep_copy(self) -> "DictState":
        """
        Creates a deep copy of the current state.
        """
        return DictState(state_data=deepcopy(self.state_data))

    @model_validator(mode="after")
    def _ensure_deep_copy(self):
        # Ensures that state_data is always a deep copy
        self.state_data = deepcopy(self.state_data)
        return self



import yaml
from typing import List, Union, Any, Literal
from swarmauri_standard.documents.Document import Document
from swarmauri_base.parsers.ParserBase import ParserBase
from swarmauri_core.ComponentBase import ComponentBase

@ComponentBase.register_type(ParserBase, 'OpenAPISpecParser')
class OpenAPISpecParser(ParserBase):
    """
    A parser that processes OpenAPI Specification files (YAML or JSON format)
    and extracts information into structured Document instances. 
    This is useful for building documentation, APIs inventory, or analyzing the API specification.
    """
    type: Literal['OpenAPISpecParser'] = 'OpenAPISpecParser'
    
    def parse(self, data: Union[str, Any]) -> List[Document]:
        """
        Parses an OpenAPI Specification from a YAML or JSON string into a list of Document instances.

        Parameters:
        - data (Union[str, Any]): The OpenAPI specification in YAML or JSON format as a string.

        Returns:
        - List[IDocument]: A list of Document instances representing the parsed information.
        """
        try:
            # Load the OpenAPI spec into a Python dictionary
            spec_dict = yaml.safe_load(data)
        except yaml.YAMLError as e:
            raise ValueError(f"Failed to parse the OpenAPI specification: {e}")
        
        documents = []
        # Iterate over paths in the OpenAPI spec to extract endpoint information
        for path, path_item in spec_dict.get("paths", {}).items():
            for method, operation in path_item.items():
                # Create a Document instance for each operation
                content = yaml.dump(operation)
                metadata = {
                    "path": path,
                    "method": method.upper(),
                    "operationId": operation.get("operationId", "")
                }
                document = Document(content=content, metadata=metadata)
                documents.append(document)

        return documents


from typing import Literal
from swarmauri_base.prompts.PromptGeneratorBase import PromptGeneratorBase
from swarmauri_core.ComponentBase import ComponentBase

@ComponentBase.register_type(PromptGeneratorBase, 'PromptGenerator')
class PromptGenerator(PromptGeneratorBase):
    type: Literal['PromptGenerator'] = 'PromptGenerator'


from typing import  Dict, Any, Literal

from swarmauri_base.tools.ToolBase import ToolBase
from swarmauri_base.schema_converters.SchemaConverterBase import SchemaConverterBase
from swarmauri_core.ComponentBase import ComponentBase, SubclassUnion

@ComponentBase.register_type(SchemaConverterBase, 'AnthropicSchemaConverter')
class AnthropicSchemaConverter(SchemaConverterBase):
    type: Literal['AnthropicSchemaConverter'] = 'AnthropicSchemaConverter'

    def convert(self, tool: SubclassUnion[ToolBase]) -> Dict[str, Any]:
        properties = {}
        required = []

        for param in tool.parameters:
            properties[param.name] = {
                "type": param.type,
                "description": param.description,
            }
            if param.required:
                required.append(param.name)

        return {
            "name": tool.name,
            "description": tool.description,
            "input_schema": {
                "type": "object",
                "properties": properties,
                "required": required,
            }
        }



import re
import base64


def encode_file(file_path):
    with open(file_path, "rb") as file:
        return base64.b64encode(file.read()).decode("utf-8")


def is_url(string):
    # Regular expression pattern to match a typical URL structure
    url_pattern = re.compile(
        r"^(https?://)?"  # Optional http or https
        r"([a-zA-Z0-9-]+\.)+[a-zA-Z]{2,6}"  # Domain name
        r"(/[a-zA-Z0-9-._~:/?#[\]@!$&\'()*+,;=%]*)?$"  # Optional path
    )
    return re.match(url_pattern, string) is not None



from typing import List, Union, Any, Literal
from swarmauri_base.chunkers.ChunkerBase import ChunkerBase
from swarmauri_core.ComponentBase import ComponentBase

@ComponentBase.register_type(ChunkerBase, 'FixedLengthChunker')
class FixedLengthChunker(ChunkerBase):
    """
    Concrete implementation of ChunkerBase that divides text into fixed-length chunks.
    
    This chunker breaks the input text into chunks of a specified size, making sure 
    that each chunk has exactly the number of characters specified by the chunk size, 
    except for possibly the last chunk.
    """
    chunk_size: int = 256
    type: Literal['FixedLengthChunker'] = 'FixedLengthChunker'
    
    def chunk_text(self, text: Union[str, Any], *args, **kwargs) -> List[str]:
        """
        Splits the input text into fixed-length chunks.

        Parameters:
        - text (Union[str, Any]): The input text to be chunked.
        
        Returns:
        - List[str]: A list of text chunks, each of a specified fixed length.
        """
        # Check if the input is a string, if not, attempt to convert to a string.
        if not isinstance(text, str):
            text = str(text)
        
        # Using list comprehension to split text into chunks of fixed size
        chunks = [text[i:i+self.chunk_size] for i in range(0, len(text), self.chunk_size)]
        
        return chunks


import json
import os
import sqlite3
import tempfile
from typing import List, Optional, Literal, Dict
import numpy as np
from swarmauri_standard.vectors.Vector import Vector
from swarmauri_standard.documents.Document import Document
from swarmauri_standard.distances.CosineDistance import CosineDistance
from swarmauri_base.vector_stores.VectorStoreBase import VectorStoreBase
from swarmauri_base.vector_stores.VectorStoreRetrieveMixin import (
    VectorStoreRetrieveMixin,
)
from swarmauri_base.vector_stores.VectorStoreSaveLoadMixin import (
    VectorStoreSaveLoadMixin,
)

from swarmauri_core.ComponentBase import ComponentBase
@ComponentBase.register_type(VectorStoreBase, 'SqliteVectorStore')
class SqliteVectorStore(
    VectorStoreSaveLoadMixin, VectorStoreRetrieveMixin, VectorStoreBase
):
    type: Literal["SqliteVectorStore"] = "SqliteVectorStore"
    db_path: str = tempfile.NamedTemporaryFile(suffix=".db", delete=False).name

    def __init__(self, db_path: Optional[str] = None, **kwargs):
        super().__init__(**kwargs)
        self._distance = CosineDistance()
        self.documents: List[Document] = []
        if db_path is not None:
            self.db_path = db_path

        # Create the SQLite database and table if they don't exist
        self._create_table()

    def _create_table(self):
        conn = sqlite3.connect(self.db_path)
        c = conn.cursor()

        # Create the documents table
        c.execute(
            """CREATE TABLE IF NOT EXISTS documents
                     (id TEXT PRIMARY KEY,
                      content TEXT,
                      metadata TEXT,
                      embedding BLOB)"""
        )

        conn.commit()
        conn.close()

    def add_document(self, document: Document) -> None:
        self.documents.append(document)
        self._insert_document(document)

    def _insert_document(self, document: Document):
        conn = sqlite3.connect(self.db_path)
        c = conn.cursor()

        # Serialize metadata to JSON
        metadata_json = json.dumps(document.metadata)

        # Serialize embedding (specifically the value of the Vector)
        embedding_data = {
            "value": document.embedding.value  # Assuming embedding is of type Vector and has a 'value' attribute
        }
        embedding_json = json.dumps(embedding_data)

        # Insert the document into the documents table
        c.execute(
            "INSERT INTO documents (id, content, metadata, embedding) VALUES (?, ?, ?, ?)",
            (document.id, document.content, metadata_json, embedding_json),
        )
        conn.commit()
        conn.close()

    def add_documents(self, documents: List[Document]) -> None:
        self.documents.extend(documents)
        for document in documents:
            self._insert_document(document)

    def get_document(self, document_id: str) -> Document:
        conn = sqlite3.connect(self.db_path)
        c = conn.cursor()

        c.execute(
            "SELECT id, content, metadata, embedding FROM documents WHERE id = ?",
            (document_id,),
        )
        row = c.fetchone()

        conn.close()

        if row:
            document_id, content, metadata_json, embedding_json = row
            metadata = json.loads(metadata_json)
            embedding_data = json.loads(embedding_json)
            embedding = Vector(
                value=embedding_data["value"]
            )  # Reconstruct the Vector object

            return Document(
                id=document_id, content=content, metadata=metadata, embedding=embedding
            )
        return None

    def get_all_documents(self) -> List[Document]:
        conn = sqlite3.connect(self.db_path)
        c = conn.cursor()

        c.execute("SELECT * FROM documents")
        rows = c.fetchall()

        conn.close()

        return [
            Document(id=row[0], content=row[1], metadata=row[2], embedding=row[3])
            for row in rows
        ]

    def delete_document(self, document_id: str) -> None:
        conn = sqlite3.connect(self.db_path)
        c = conn.cursor()

        c.execute("DELETE FROM documents WHERE id = ?", (document_id,))

        conn.commit()
        conn.close()

    def update_document(self, document: Document) -> None:
        conn = sqlite3.connect(self.db_path)
        c = conn.cursor()

        c.execute(
            """UPDATE documents 
               SET content = ?, metadata = ?, embedding = ? 
               WHERE id = ?""",
            (document.content, document.metadata, document.embedding, document.id),
        )

        conn.commit()
        conn.close()

    def retrieve(self, query_vector: List[float], top_k: int = 5) -> List[Dict]:
        conn = sqlite3.connect(self.db_path)
        c = conn.cursor()
        c.execute("SELECT id, embedding FROM documents")
        rows = c.fetchall()
        conn.close()

        # Convert query vector to numpy array
        query_vector = np.array(query_vector, dtype=np.float32)

        results = []
        for row in rows:
            doc_id, embedding_json = row
            embedding = json.loads(embedding_json)
            vector = np.array(embedding["value"], dtype=np.float32)

            # Compute similarity (e.g., Euclidean distance)
            distance = np.linalg.norm(query_vector - vector)
            results.append((doc_id, distance))

        # Sort results by distance
        results.sort(key=lambda x: x[1])

        # Get top_k results
        top_results = [doc_id for doc_id, _ in results[:top_k]]
        return top_results

    def close(self):
        # Clean up the database file
        if os.path.exists(self.db_path):
            os.remove(self.db_path)



from typing import Literal
from swarmauri_base.vectors.VectorBase import VectorBase
from swarmauri_core.ComponentBase import ComponentBase

@ComponentBase.register_type(VectorBase, 'Vector')
class Vector(VectorBase):
    type: Literal['Vector'] = 'Vector'


import yake
from typing import List, Union, Any, Literal
from pydantic import ConfigDict, PrivateAttr
from swarmauri_standard.documents.Document import Document
from swarmauri_base.parsers.ParserBase import ParserBase
from swarmauri_core.ComponentBase import ComponentBase

@ComponentBase.register_type(ParserBase, 'KeywordExtractorParser')
class KeywordExtractorParser(ParserBase):
    """
    Extracts keywords from text using the YAKE keyword extraction library.
    """
    lang: str = 'en'
    num_keywords: int = 10
    _kw_extractor: yake.KeywordExtractor = PrivateAttr(default=None)
    model_config = ConfigDict(extra='forbid', arbitrary_types_allowed=True)
    type: Literal['KeywordExtractorParser'] = 'KeywordExtractorParser'
    
    def __init__(self, **data):
        super().__init__(**data)
        self._kw_extractor = yake.KeywordExtractor(lan=self.lang,
                                                   n=3, 
                                                   dedupLim=0.9, 
                                                   dedupFunc='seqm', 
                                                   windowsSize=1, 
                                                   top=self.num_keywords, 
                                                   features=None)
    

    def parse(self, data: Union[str, Any]) -> List[Document]:
        """
        Extract keywords from input text and return as list of Document instances containing keyword information.

        Parameters:
        - data (Union[str, Any]): The input text from which to extract keywords.

        Returns:
        - List[Document]: A list of Document instances, each containing information about an extracted keyword.
        """
        # Ensure data is in string format for analysis
        text = str(data) if not isinstance(data, str) else data

        # Extract keywords using YAKE
        keywords = self._kw_extractor.extract_keywords(text)

        # Create Document instances for each keyword
        documents = [Document(content=keyword, metadata={"score": score}) for index, (keyword, score) in enumerate(keywords)]
        
        return documents


from typing import Literal
from pydantic import Field

from swarmauri_base.toolkits.ToolkitBase import ToolkitBase

from swarmauri_standard.tools.concrete import (
    AutomatedReadabilityIndexTool,
    ColemanLiauIndexTool,
    FleschKincaidTool,
    FleschReadingEaseTool,
    GunningFogTool,
)

from swarmauri_core.ComponentBase import ComponentBase

@ComponentBase.register_type(ToolkitBase, 'AccessibilityToolkit')
class AccessibilityToolkit(ToolkitBase):
    type: Literal["AccessibilityToolkit"] = "AccessibilityToolkit"

    resource: str = "AccessibilityToolkit"

    # Explicitly define the tools as fields
    automated_readability_index_tool: AutomatedReadabilityIndexTool = Field(
        default_factory=lambda: AutomatedReadabilityIndexTool(
            name="AutomatedReadabilityIndexTool"
        )
    )
    coleman_liau_index_tool: ColemanLiauIndexTool = Field(
        default_factory=lambda: ColemanLiauIndexTool(name="ColemanLiauIndexTool")
    )
    flesch_kincaid_tool: FleschKincaidTool = Field(
        default_factory=lambda: FleschKincaidTool(name="FleschKincaidTool")
    )
    flesch_reading_ease_tool: FleschReadingEaseTool = Field(
        default_factory=lambda: FleschReadingEaseTool(name="FleschReadingEaseTool")
    )
    gunning_fog_tool: GunningFogTool = Field(
        default_factory=lambda: GunningFogTool(name="GunningFogTool")
    )

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # Add all tools to the toolkit using add_tools method
        self.add_tool(self.automated_readability_index_tool)
        self.add_tool(self.coleman_liau_index_tool)
        self.add_tool(self.flesch_kincaid_tool)
        self.add_tool(self.flesch_reading_ease_tool)
        self.add_tool(self.gunning_fog_tool)



from typing import List, Union, Any, Optional, Literal
import re
from swarmauri_base.chunkers.ChunkerBase import ChunkerBase
from swarmauri_core.ComponentBase import ComponentBase

@ComponentBase.register_type(ChunkerBase, 'MdSnippetChunker')
class MdSnippetChunker(ChunkerBase):
    version: str = "0.1.0.dev1"
    language: Optional[str] = None
    type: Literal['MdSnippetChunker'] = 'MdSnippetChunker'
    def chunk_text(self, text: Union[str, Any], *args, **kwargs) -> List[tuple]:
        """
        Extracts paired comments and code blocks from Markdown content based on the 
        specified programming language.
        """
        if self.language:
            # If language is specified, directly extract the corresponding blocks
            pattern = fr'```{self.language}\s*(.*?)```'
            scripts = re.findall(pattern, text, re.DOTALL)
            comments_temp = re.split(pattern, text, flags=re.DOTALL)
        else:
            # Extract blocks and languages dynamically if no specific language is provided
            scripts = []
            languages = []
            for match in re.finditer(r'```(\w+)?\s*(.*?)```', text, re.DOTALL):
                if match.group(1) is not None:  # Checks if a language identifier is present
                    languages.append(match.group(1))
                    scripts.append(match.group(2))
                else:
                    languages.append('')  # Default to an empty string if no language is found
                    scripts.append(match.group(2))
            comments_temp = re.split(r'```.*?```', text, flags=re.DOTALL)

        comments = [comment.strip() for comment in comments_temp]

        if text.strip().startswith('```'):
            comments = [''] + comments
        if text.strip().endswith('```'):
            comments.append('')

        if self.language:
            structured_output = [(comments[i], self.language, scripts[i].strip()) for i in range(len(scripts))]
        else:
            structured_output = [(comments[i], languages[i], scripts[i].strip()) for i in range(len(scripts))]

        return structured_output



import httpx
import time
import asyncio
import contextlib
from typing import List, Literal, Optional, Dict
from pydantic import PrivateAttr
from swarmauri_standard.utils.retry_decorator import retry_on_status_codes
from swarmauri_base.image_gens.ImageGenBase import ImageGenBase
from swarmauri_core.ComponentBase import ComponentBase

@ComponentBase.register_type(ImageGenBase, 'BlackForestImgGenModel')
class BlackForestImgGenModel(ImageGenBase):
    """
    A model for generating images using FluxPro's image generation models through the Black Forest API.
    Link to API key: https://api.bfl.ml/auth/profile
    """

    _BASE_URL: str = PrivateAttr("https://api.bfl.ml")
    _client: httpx.Client = PrivateAttr(default=None)
    _async_client: httpx.AsyncClient = PrivateAttr(default=None)
    _headers: Dict[str, str] = PrivateAttr(default=None)

    api_key: str
    allowed_models: List[str] = ["flux-pro-1.1", "flux-pro", "flux-dev"]

    name: str = "flux-pro"  # Default model
    type: Literal["BlackForestImgGenModel"] = "BlackForestImgGenModel"

    def __init__(self, **kwargs):
        """
        Initializes the BlackForestImgGenModel instance with HTTP clients.
        """
        super().__init__(**kwargs)
        self._headers = {
            "Content-Type": "application/json",
            "X-Key": self.api_key,
        }
        self._client = httpx.Client(headers=self._headers, timeout=30)

    async def _get_async_client(self) -> httpx.AsyncClient:
        """Gets or creates an async client instance."""
        if self._async_client is None or self._async_client.is_closed:
            self._async_client = httpx.AsyncClient(headers=self._headers, timeout=30)
        return self._async_client

    async def _close_async_client(self):
        """Closes the async client if it exists and is open."""
        if self._async_client is not None and not self._async_client.is_closed:
            await self._async_client.aclose()
            self._async_client = None

    @retry_on_status_codes((429, 529), max_retries=1)
    def _send_request(self, endpoint: str, data: dict) -> dict:
        """Send a synchronous request to FluxPro's API for image generation."""
        url = f"{self._BASE_URL}/{endpoint}"
        response = self._client.post(url, json=data)
        response.raise_for_status()
        return response.json()

    @retry_on_status_codes((429, 529), max_retries=1)
    async def _async_send_request(self, endpoint: str, data: dict) -> dict:
        """Send an asynchronous request to FluxPro's API for image generation."""
        client = await self._get_async_client()
        url = f"{self._BASE_URL}/{endpoint}"
        response = await client.post(url, json=data)
        response.raise_for_status()
        return response.json()

    @retry_on_status_codes((429, 529), max_retries=1)
    def _get_result(self, task_id: str) -> dict:
        """Get the result of a generation task synchronously."""
        url = f"{self._BASE_URL}/v1/get_result"
        params = {"id": task_id}
        response = self._client.get(url, params=params)
        response.raise_for_status()
        return response.json()

    @retry_on_status_codes((429, 529), max_retries=1)
    async def _async_get_result(self, task_id: str) -> dict:
        """Get the result of a generation task asynchronously."""
        client = await self._get_async_client()
        url = f"{self._BASE_URL}/v1/get_result"
        params = {"id": task_id}
        response = await client.get(url, params=params)
        response.raise_for_status()
        return response.json()

    def generate_image(
        self,
        prompt: str,
        width: int = 1024,
        height: int = 768,
        steps: Optional[int] = None,
        prompt_upsampling: bool = False,
        seed: Optional[int] = None,
        guidance: Optional[float] = None,
        safety_tolerance: Optional[int] = None,
        interval: Optional[float] = None,
        max_wait_time: int = 300,
        check_interval: int = 10,
    ) -> Dict:
        """
        Generates an image based on the prompt and waits for the result synchronously.

        Args:
            prompt (str): The text prompt for image generation
            width (int): Image width in pixels
            height (int): Image height in pixels
            steps (Optional[int]): Number of inference steps
            prompt_upsampling (bool): Whether to use prompt upsampling
            seed (Optional[int]): Random seed for generation
            guidance (Optional[float]): Guidance scale
            safety_tolerance (Optional[int]): Safety tolerance level
            interval (Optional[float]): Interval parameter (flux-pro only)
            max_wait_time (int): Maximum time to wait for result in seconds
            check_interval (int): Time between status checks in seconds

        Returns:
            Dict: Dictionary containing the image URL and other result information
        """
        endpoint = f"v1/{self.name}"
        data = {
            "prompt": prompt,
            "width": width,
            "height": height,
            "prompt_upsampling": prompt_upsampling,
        }

        if steps is not None:
            data["steps"] = steps
        if seed is not None:
            data["seed"] = seed
        if guidance is not None:
            data["guidance"] = guidance
        if safety_tolerance is not None:
            data["safety_tolerance"] = safety_tolerance
        if interval is not None and self.name == "flux-pro":
            data["interval"] = interval

        response = self._send_request(endpoint, data)
        task_id = response["id"]

        start_time = time.time()
        while time.time() - start_time < max_wait_time:
            result = self._get_result(task_id)
            if result["status"] == "Ready":
                return result["result"]["sample"]
            elif result["status"] in [
                "Error",
                "Request Moderated",
                "Content Moderated",
            ]:
                raise Exception(f"Task failed with status: {result['status']}")
            time.sleep(check_interval)

        raise TimeoutError(f"Image generation timed out after {max_wait_time} seconds")

    async def agenerate_image(self, prompt: str, **kwargs) -> Dict:
        """
        Asynchronously generates an image based on the prompt and waits for the result.

        Args:
            prompt (str): The text prompt for image generation
            **kwargs: Additional arguments passed to generate_image

        Returns:
            Dict: Dictionary containing the image URL and other result information
        """
        try:
            endpoint = f"v1/{self.name}"
            data = {
                "prompt": prompt,
                "width": kwargs.get("width", 1024),
                "height": kwargs.get("height", 768),
                "prompt_upsampling": kwargs.get("prompt_upsampling", False),
            }

            optional_params = [
                "steps",
                "seed",
                "guidance",
                "safety_tolerance",
            ]
            for param in optional_params:
                if param in kwargs:
                    data[param] = kwargs[param]

            if "interval" in kwargs and self.name == "flux-pro":
                data["interval"] = kwargs["interval"]

            response = await self._async_send_request(endpoint, data)
            task_id = response["id"]

            max_wait_time = kwargs.get("max_wait_time", 300)
            check_interval = kwargs.get("check_interval", 10)
            start_time = time.time()

            while time.time() - start_time < max_wait_time:
                result = await self._async_get_result(task_id)
                if result["status"] == "Ready":
                    return result["result"]["sample"]
                elif result["status"] in [
                    "Error",
                    "Request Moderated",
                    "Content Moderated",
                ]:
                    raise Exception(f"Task failed with status: {result['status']}")
                await asyncio.sleep(check_interval)

            raise TimeoutError(
                f"Image generation timed out after {max_wait_time} seconds"
            )
        finally:
            await self._close_async_client()

    def batch_generate(self, prompts: List[str], **kwargs) -> List[Dict]:
        """
        Generates images for a batch of prompts synchronously.

        Args:
            prompts (List[str]): List of text prompts
            **kwargs: Additional arguments passed to generate_image

        Returns:
            List[Dict]: List of result dictionaries
        """
        return [self.generate_image(prompt=prompt, **kwargs) for prompt in prompts]

    async def abatch_generate(
        self, prompts: List[str], max_concurrent: int = 5, **kwargs
    ) -> List[Dict]:
        """
        Asynchronously generates images for a batch of prompts.

        Args:
            prompts (List[str]): List of text prompts
            max_concurrent (int): Maximum number of concurrent tasks
            **kwargs: Additional arguments passed to agenerate_image

        Returns:
            List[Dict]: List of result dictionaries
        """
        try:
            semaphore = asyncio.Semaphore(max_concurrent)

            async def process_prompt(prompt):
                async with semaphore:
                    return await self.agenerate_image(prompt=prompt, **kwargs)

            tasks = [process_prompt(prompt) for prompt in prompts]
            return await asyncio.gather(*tasks)
        finally:
            await self._close_async_client()

    def __del__(self):
        """Cleanup method to ensure clients are closed."""
        self._client.close()
        if self._async_client is not None and not self._async_client.is_closed:
            with contextlib.suppress(Exception):
                asyncio.run(self._close_async_client())



from datetime import datetime
import uuid
from typing import Dict, Any, Optional

from swarmauri_core.tracing.ITracer import ITracer
from swarmauri_standard.tracing.SimpleTraceContext import SimpleTraceContext

class SimpleTracer(ITracer):
    _instance = None  # Singleton instance placeholder

    @classmethod
    def instance(cls):
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance

    def __init__(self):
        if SimpleTracer._instance is not None:
            raise RuntimeError("SimpleTracer is a singleton. Use SimpleTracer.instance().")
        self.trace_stack = []

    def start_trace(self, name: str, initial_attributes: Optional[Dict[str, Any]] = None) -> SimpleTraceContext:
        trace_id = str(uuid.uuid4())
        trace_context = SimpleTraceContext(trace_id, name, initial_attributes)
        self.trace_stack.append(trace_context)
        return trace_context

    def end_trace(self):
        if self.trace_stack:
            completed_trace = self.trace_stack.pop()
            completed_trace.close()
            # Example of simply printing the completed trace; in practice, you might log it or store it elsewhere
            print(f"Trace Completed: {completed_trace.name}, Duration: {completed_trace.start_time} to {completed_trace.end_time}, Attributes: {completed_trace.attributes}")

    def annotate_trace(self, key: str, value: Any):
        if not self.trace_stack:
            print("No active trace to annotate.")
            return
        current_trace = self.trace_stack[-1]
        current_trace.add_attribute(key, value)


import asyncio
import httpx
import aiofiles

from typing import Dict, List, Literal
from pydantic import PrivateAttr

from swarmauri_standard.utils.retry_decorator import retry_on_status_codes
from swarmauri_base.llms.LLMBase import LLMBase
from swarmauri_core.ComponentBase import ComponentBase

@ComponentBase.register_type(LLMBase, 'GroqAIAudio')
class GroqAIAudio(LLMBase):
    """
    GroqAIAudio is a class that provides transcription and translation capabilities
    using Groq's audio models. It supports both synchronous and asynchronous methods
    for processing audio files.

    Attributes:
        api_key (str): API key for authentication.
        allowed_models (List[str]): List of supported model names.
        name (str): The default model name to be used for predictions.
        type (Literal["GroqAIAudio"]): The type identifier for the class.
    """

    api_key: str
    allowed_models: List[str] = [
        "distil-whisper-large-v3-en",
        "whisper-large-v3",
    ]

    name: str = "distil-whisper-large-v3-en"
    type: Literal["GroqAIAudio"] = "GroqAIAudio"
    _client: httpx.Client = PrivateAttr(default=None)
    _async_client: httpx.AsyncClient = PrivateAttr(default=None)
    _BASE_URL: str = PrivateAttr(default="https://api.groq.com/openai/v1/audio/")

    def __init__(self, **data):
        """
        Initialize the GroqAIAudio class with the provided data.

        Args:
            **data: Arbitrary keyword arguments containing initialization data.
        """
        super().__init__(**data)
        self._client = httpx.Client(
            headers={"Authorization": f"Bearer {self.api_key}"},
            base_url=self._BASE_URL,
            timeout=30,
        )
        self._async_client = httpx.AsyncClient(
            headers={"Authorization": f"Bearer {self.api_key}"},
            base_url=self._BASE_URL,
            timeout=30,
        )

    @retry_on_status_codes((429, 529), max_retries=1)
    def predict(
        self,
        audio_path: str,
        task: Literal["transcription", "translation"] = "transcription",
    ) -> str:
        """
        Perform synchronous transcription or translation on the provided audio file.

        Args:
            audio_path (str): Path to the audio file.
            task (Literal["transcription", "translation"]): Task type. Defaults to "transcription".

        Returns:
            str: The resulting transcription or translation text.

        Raises:
            ValueError: If the specified task is not supported.
            httpx.HTTPStatusError: If the API request fails.
        """
        kwargs = {
            "model": self.name,
        }

        if task == "translation":
            kwargs["model"] = "whisper-large-v3"

        with open(audio_path, "rb") as audio_file:
            actions = {
                "transcription": self._client.post(
                    "transcriptions", files={"file": audio_file}, data=kwargs
                ),
                "translation": self._client.post(
                    "translations", files={"file": audio_file}, data=kwargs
                ),
            }

            if task not in actions:
                raise ValueError(
                    f"Task {task} not supported. Choose from {list(actions)}"
                )

        response = actions[task]
        response.raise_for_status()

        response_data = response.json()

        return response_data["text"]

    @retry_on_status_codes((429, 529), max_retries=1)
    async def apredict(
        self,
        audio_path: str,
        task: Literal["transcription", "translation"] = "transcription",
    ) -> str:
        """
        Perform asynchronous transcription or translation on the provided audio file.

        Args:
            audio_path (str): Path to the audio file.
            task (Literal["transcription", "translation"]): Task type. Defaults to "transcription".

        Returns:
            str: The resulting transcription or translation text.

        Raises:
            ValueError: If the specified task is not supported.
            httpx.HTTPStatusError: If the API request fails.
        """
        kwargs = {
            "model": self.name,
        }
        if task == "translation":
            kwargs["model"] = "whisper-large-v3"

        async with aiofiles.open(audio_path, "rb") as audio_file:
            file_content = await audio_file.read()
            file_name = audio_path.split("/")[-1]
            actions = {
                "transcription": await self._async_client.post(
                    "transcriptions",
                    files={"file": (file_name, file_content, "audio/wav")},
                    data=kwargs,
                ),
                "translation": await self._async_client.post(
                    "translations",
                    files={"file": (file_name, file_content, "audio/wav")},
                    data=kwargs,
                ),
            }
            if task not in actions:
                raise ValueError(
                    f"Task {task} not supported. Choose from {list(actions)}"
                )

            response = actions[task]
            response.raise_for_status()

            response_data = response.json()
            return response_data["text"]

    def batch(
        self,
        path_task_dict: Dict[str, Literal["transcription", "translation"]],
    ) -> List:
        """
        Synchronously process multiple audio files for transcription or translation.

        Args:
            path_task_dict (Dict[str, Literal["transcription", "translation"]]): A dictionary where
                the keys are paths to audio files and the values are the tasks.

        Returns:
            List: A list of resulting texts from each audio file.
        """
        return [
            self.predict(audio_path=path, task=task)
            for path, task in path_task_dict.items()
        ]

    async def abatch(
        self,
        path_task_dict: Dict[str, Literal["transcription", "translation"]],
        max_concurrent=5,
    ) -> List:
        """
        Asynchronously process multiple audio files for transcription or translation
        with controlled concurrency.

        Args:
            path_task_dict (Dict[str, Literal["transcription", "translation"]]): A dictionary where
                the keys are paths to audio files and the values are the tasks.
            max_concurrent (int): Maximum number of concurrent tasks. Defaults to 5.

        Returns:
            List: A list of resulting texts from each audio file.
        """
        semaphore = asyncio.Semaphore(max_concurrent)

        async def process_conversation(path, task) -> str:
            async with semaphore:
                return await self.apredict(audio_path=path, task=task)

        tasks = [
            process_conversation(path, task) for path, task in path_task_dict.items()
        ]
        return await asyncio.gather(*tasks)



from typing import Literal
from swarmauri_base.prompts.PromptTemplateBase import PromptTemplateBase
from swarmauri_core.ComponentBase import ComponentBase

@ComponentBase.register_type(PromptTemplateBase, 'PromptTemplate')
class PromptTemplate(PromptTemplateBase):
    type: Literal['PromptTemplate'] = 'PromptTemplate'


# from swarmauri_standard.utils._lazy_import import _lazy_import

# # List of chunker names (file names without the ".py" extension) and corresponding class names
# chunkers_files = [
#     ("swarmauri_standard.chunkers.DelimiterBasedChunker", "DelimiterBasedChunker"),
#     ("swarmauri_standard.chunkers.FixedLengthChunker", "FixedLengthChunker"),
#     ("swarmauri_standard.chunkers.MdSnippetChunker", "MdSnippetChunker"),
#     ("swarmauri_standard.chunkers.SentenceChunker", "SentenceChunker"),
#     ("swarmauri_standard.chunkers.SlidingWindowChunker", "SlidingWindowChunker"),
# ]

# # Lazy loading of chunker classes, storing them in variables
# for module_name, class_name in chunkers_files:
#     globals()[class_name] = _lazy_import(module_name, class_name)

# # Adding the lazy-loaded chunker classes to __all__
# __all__ = [class_name for _, class_name in chunkers_files]




# from swarmauri_standard.utils._lazy_import import _lazy_import

# # List of vectore_stores names (file names without the ".py" extension) and corresponding class names
# vectore_stores_files = [
#     ("swarmauri_standard.vector_stores.SqliteVectorStore", "SqliteVectorStore"),
#     ("swarmauri_standard.vector_stores.TfidfVectorStore", "TfidfVectorStore"),
# ]

# # Lazy loading of vectore_storess, storing them in variables
# for module_name, class_name in vectore_stores_files:
#     globals()[class_name] = _lazy_import(module_name, class_name)

# # Adding the lazy-loaded vectore_storess to __all__
# __all__ = [class_name for _, class_name in vectore_stores_files]



# # swarmauri_standard/llms/__init__.py

# from typing import TYPE_CHECKING

# from swarmauri_standard.utils.LazyLoader import LazyLoader  # Ensure correct import path

# # List of LLMs (Large Language Models) names (module paths) and corresponding class names
# llms_files = [
#     ("swarmauri_standard.llms.AI21StudioModel", "AI21StudioModel"),
#     ("swarmauri_standard.llms.AnthropicModel", "AnthropicModel"),
#     ("swarmauri_standard.llms.AnthropicToolModel", "AnthropicToolModel"),
#     ("swarmauri_standard.llms.BlackForestImgGenModel", "BlackForestImgGenModel"),
#     ("swarmauri_standard.llms.CohereModel", "CohereModel"),
#     ("swarmauri_standard.llms.CohereToolModel", "CohereToolModel"),
#     ("swarmauri_standard.llms.DeepInfraImgGenModel", "DeepInfraImgGenModel"),
#     ("swarmauri_standard.llms.DeepInfraModel", "DeepInfraModel"),
#     ("swarmauri_standard.llms.DeepSeekModel", "DeepSeekModel"),
#     ("swarmauri_standard.llms.FalAIImgGenModel", "FalaiImgGenModel"),
#     ("swarmauri_standard.llms.FalAIVisionModel", "FalAIVisionModel"),
#     ("swarmauri_standard.llms.GeminiProModel", "GeminiProModel"),
#     ("swarmauri_standard.llms.GeminiToolModel", "GeminiToolModel"),
#     ("swarmauri_standard.llms.GroqAIAudio", "GroqAIAudio"),
#     ("swarmauri_standard.llms.GroqModel", "GroqModel"),
#     ("swarmauri_standard.llms.GroqToolModel", "GroqToolModel"),
#     ("swarmauri_standard.llms.GroqVisionModel", "GroqVisionModel"),
#     ("swarmauri_standard.llms.HyperbolicAudioTTS", "HyperbolicAudioTTS"),
#     ("swarmauri_standard.llms.HyperbolicImgGenModel", "HyperbolicImgGenModel"),
#     ("swarmauri_standard.llms.HyperbolicModel", "HyperbolicModel"),
#     ("swarmauri_standard.llms.HyperbolicVisionModel", "HyperbolicVisionModel"),
#     ("swarmauri_standard.llms.MistralModel", "MistralModel"),
#     ("swarmauri_standard.llms.MistralToolModel", "MistralToolModel"),
#     ("swarmauri_standard.llms.OpenAIAudio", "OpenAIAudio"),
#     ("swarmauri_standard.llms.OpenAIAudioTTS", "OpenAIAudioTTS"),
#     ("swarmauri_standard.llms.OpenAIImgGenModel", "OpenAIImgGenModel"),
#     ("swarmauri_standard.llms.OpenAIModel", "OpenAIModel"),
#     ("swarmauri_standard.llms.OpenAIToolModel", "OpenAIToolModel"),
#     ("swarmauri_standard.llms.PerplexityModel", "PerplexityModel"),
#     ("swarmauri_standard.llms.PlayHTModel", "PlayHTModel"),
#     ("swarmauri_standard.llms.WhisperLargeModel", "WhisperLargeModel"),
# ]

# # Initialize LazyLoader instances for each LLM class
# for module_path, class_name in llms_files:
#     globals()[class_name] = LazyLoader(module_path, class_name)

# # Define __all__ to include all lazy-loaded classes
# __all__ = [class_name for _, class_name in llms_files]

# # Type hinting for static analysis and IDE support
# if TYPE_CHECKING:
#     from swarmauri_standard.llms.AI21StudioModel import AI21StudioModel
#     from swarmauri_standard.llms.AnthropicModel import AnthropicModel
#     from swarmauri_standard.llms.AnthropicToolModel import AnthropicToolModel
#     from swarmauri_standard.llms.BlackForestImgGenModel import BlackForestImgGenModel
#     from swarmauri_standard.llms.CohereModel import CohereModel
#     from swarmauri_standard.llms.CohereToolModel import CohereToolModel
#     from swarmauri_standard.llms.DeepInfraImgGenModel import DeepInfraImgGenModel
#     from swarmauri_standard.llms.DeepInfraModel import DeepInfraModel
#     from swarmauri_standard.llms.DeepSeekModel import DeepSeekModel
#     from swarmauri_standard.llms.FalAIImgGenModel import FalaiImgGenModel
#     from swarmauri_standard.llms.FalAIVisionModel import FalAIVisionModel
#     from swarmauri_standard.llms.GeminiProModel import GeminiProModel
#     from swarmauri_standard.llms.GeminiToolModel import GeminiToolModel
#     from swarmauri_standard.llms.GroqAIAudio import GroqAIAudio
#     from swarmauri_standard.llms.GroqModel import GroqModel
#     from swarmauri_standard.llms.GroqToolModel import GroqToolModel
#     from swarmauri_standard.llms.GroqVisionModel import GroqVisionModel
#     from swarmauri_standard.llms.HyperbolicAudioTTS import HyperbolicAudioTTS
#     from swarmauri_standard.llms.HyperbolicImgGenModel import HyperbolicImgGenModel
#     from swarmauri_standard.llms.HyperbolicModel import HyperbolicModel
#     from swarmauri_standard.llms.HyperbolicVisionModel import HyperbolicVisionModel
#     from swarmauri_standard.llms.MistralModel import MistralModel
#     from swarmauri_standard.llms.MistralToolModel import MistralToolModel
#     from swarmauri_standard.llms.OpenAIAudio import OpenAIAudio
#     from swarmauri_standard.llms.OpenAIAudioTTS import OpenAIAudioTTS
#     from swarmauri_standard.llms.OpenAIImgGenModel import OpenAIImgGenModel
#     from swarmauri_standard.llms.OpenAIModel import OpenAIModel
#     from swarmauri_standard.llms.OpenAIToolModel import OpenAIToolModel
#     from swarmauri_standard.llms.PerplexityModel import PerplexityModel
#     from swarmauri_standard.llms.PlayHTModel import PlayHTModel
#     from swarmauri_standard.llms.WhisperLargeModel import WhisperLargeModel



import re
from typing import Dict, List, Literal
from swarmauri_standard.tools.Parameter import Parameter
from swarmauri_base.tools.ToolBase import ToolBase
from swarmauri_core.ComponentBase import ComponentBase

@ComponentBase.register_type(ToolBase, 'AutomatedReadabilityIndexTool')
class AutomatedReadabilityIndexTool(ToolBase):
    """
    A tool for calculating the Automated Readability Index (ARI).

    Attributes:
        version (str): The version of the tool.
        name (str): The name of the tool.
        type (Literal["AutomatedReadabilityIndexTool"]): The type of the tool.
        description (str): A brief description of what the tool does.
        parameters (List[Parameter]): The parameters for configuring the tool.
    """

    version: str = "0.1.dev1"
    name: str = "AutomatedReadabilityIndexTool"
    type: Literal["AutomatedReadabilityIndexTool"] = "AutomatedReadabilityIndexTool"
    description: str = (
        "Calculates the Automated Readability Index (ARI) for a given text."
    )
    parameters: List[Parameter] = [
        Parameter(
            name="input_text",
            type="string",
            description="The input text for which to calculate the ARI.",
            required=True,
        )
    ]

    def __call__(self, input_text: str) -> Dict[str, str]:
        """
        Executes the ARI tool and returns the readability score.

        ARI formula:
        4.71 * (characters/words) + 0.5 * (words/sentences) - 21.43

        Parameters:
            input_text (str): The input text for which to calculate the ARI.

        Returns:
            str: The Automated Readability Index as a string.

        Raises:
            ValueError: If the input data is invalid.
        """
        if self.validate_input(input_text):
            text = input_text
            num_sentences = self.count_sentences(text)
            num_words = self.count_words(text)
            num_characters = self.count_characters(text)
            if num_sentences == 0 or num_words == 0:
                return {"ari_score": 0.0}
            characters_per_word = num_characters / num_words
            words_per_sentence = num_words / num_sentences
            ari_score = 4.71 * characters_per_word + 0.5 * words_per_sentence - 21.43
            return {"ari_score": ari_score}
        else:
            raise ValueError("Invalid input for AutomatedReadabilityIndexTool.")

    def count_sentences(self, text: str) -> int:
        """
        Counts the number of sentences in the text.

        Parameters:
            text (str): The input text.

        Returns:
            int: The number of sentences in the text.
        """
        sentence_endings = re.compile(r"[.!?]")
        sentences = sentence_endings.split(text)
        return len([s for s in sentences if s.strip()])  # Count non-empty sentences

    def count_words(self, text: str) -> int:
        """
        Counts the number of words in the text.

        Parameters:
            text (str): The input text.

        Returns:
            int: The number of words in the text.
        """
        words = re.findall(r"\b\w+\b", text)
        return len(words)

    def count_characters(self, text: str) -> int:
        """
        Counts the number of characters in the text.

        Parameters:
            text (str): The input text.

        Returns:
            int: The number of characters in the text, excluding spaces.
        """
        return len(text) - text.count(" ")  # Count characters excluding spaces

    def validate_input(self, input_text: str) -> bool:
        """
        Validates the input text for the ARI tool.

        Parameters:
            input_text (str): The input text to validate.

        Returns:
            bool: True if the input is valid, False otherwise.
        """
        if isinstance(input_text, str):
            return True
        return False


SubclassUnion.update(
    baseclass=ToolBase,
    type_name="AutomatedReadabilityIndexTool",
    obj=AutomatedReadabilityIndexTool,
)



import logging
from urllib.parse import urlencode
import httpx
import base64
import json
from typing import List
from swarmauri_base.documents.DocumentBase import DocumentBase
from swarmauri_base.dataconnectors.DataConnectorBase import DataConnectorBase


class GoogleDriveDataConnector(DataConnectorBase):
    """
    Data connector for interacting with Google Drive files and converting them to Swarmauri documents.

    Supports authentication, data fetching, and basic CRUD operations for Google Drive resources.
    """

    def __init__(self, credentials_path: str = None):
        """
        Initialize the Google Drive Data Connector.

        :param credentials_path: Path to the Google OAuth2 credentials JSON file
        """
        with open(credentials_path, "r") as cred_file:
            credentials = json.load(cred_file)

        self.client_id = credentials.get("client_id")
        self.client_secret = credentials.get("client_secret")
        self.redirect_uri = credentials.get("redirect_uri")

        # Tokens will be stored here
        self.access_token = None
        self.refresh_token = None

        self.authorization_code = None

        self.client = httpx.Client()

    def generate_authorization_url(self) -> str:
        """Generate the authorization URL for user consent"""
        params = {
            "client_id": self.client_id,
            "redirect_uri": self.redirect_uri,
            "response_type": "code",
            "scope": "https://www.googleapis.com/auth/drive",
            "access_type": "offline",  # This ensures we get a refresh token
        }
        return f"https://accounts.google.com/o/oauth2/v2/auth?{urlencode(params)}"

    def _exchange_code_for_tokens(self):
        """Exchange authorization code for access and refresh tokens"""
        if not self.authorization_code:
            raise ValueError("No authorization code available")

        token_url = "https://oauth2.googleapis.com/token"
        payload = {
            "client_id": self.client_id,
            "client_secret": self.client_secret,
            "code": self.authorization_code,
            "grant_type": "authorization_code",
            "redirect_uri": self.redirect_uri,
        }

        response = self.client.post(token_url, data=payload)
        tokens = response.json()

        logging.info(f"Token response: {tokens}")
        if "access_token" not in tokens:
            raise ValueError("Failed to obtain access token")
        self.access_token = tokens["access_token"]
        self.refresh_token = tokens.get("refresh_token")

    def refresh_access_token(self):
        """Refresh the access token using the refresh token"""
        if not self.refresh_token:
            raise ValueError("No refresh token available")

        token_url = "https://oauth2.googleapis.com/token"
        payload = {
            "client_id": self.client_id,
            "client_secret": self.client_secret,
            "refresh_token": self.refresh_token,
            "grant_type": "refresh_token",
        }

        response = self.client.post(token_url, data=payload)
        tokens = response.json()
        self.access_token = tokens["access_token"]

    def authenticate(self):
        """
        Authenticate with Google Drive using OAuth2.

        This method generates an authorization URL, prompts the user to visit the URL
        and enter the authorization code, and then exchanges the code for tokens.
        """
        try:
            # Generate authorization URL
            auth_url = self.generate_authorization_url()
            print("Please visit the following URL to authenticate:")
            print(auth_url)

            # Prompt for authorization code
            while True:
                authorization_code = input("Enter the authorization code: ").strip()

                if not authorization_code:
                    print("Authorization code cannot be empty. Please try again.")
                    continue

                self.authorization_code = authorization_code

                try:
                    self._exchange_code_for_tokens()
                    logging.info("Successfully authenticated and obtained tokens")
                    return
                except ValueError as e:
                    print(f"Error exchanging authorization code: {e}")
                    print("Please try again.")
                    self.authorization_code = None

        except Exception as e:
            logging.error(f"Authentication failed: {e}")
            raise ValueError(f"Authentication failed: {e}")

    def fetch_data(self, query: str = None, **kwargs) -> List[DocumentBase]:
        """
        Fetch documents from Google Drive based on a query.

        :param query: Search query for files (optional)
        :param kwargs: Additional parameters like mime_type, max_results
        :return: List of Swarmauri Documents
        """
        if not self.access_token:
            raise ValueError("Not authenticated. Call authenticate() first.")

        try:
            # Prepare request parameters
            query_str = query or ""
            mime_type = kwargs.get("mime_type", "application/vnd.google-apps.document")
            max_results = kwargs.get("max_results", 100)

            # Construct request headers and parameters
            headers = {
                "Authorization": f"Bearer {self.access_token}",
                "Accept": "application/json",
            }

            params = {
                "q": f"mimeType='{mime_type}' and name contains '{query_str}'",
                "pageSize": max_results,
                "fields": "files(id,name,mimeType)",
            }

            # Make request to Google Drive API
            response = self.client.get(
                "https://www.googleapis.com/drive/v3/files",
                headers=headers,
                params=params,
            )
            response.raise_for_status()

            files = response.json().get("files", [])

            # Convert Google Drive files to Swarmauri Documents
            documents = []
            for file in files:
                content = self._get_file_content(file["id"])
                document = DocumentBase(
                    content=content,
                    metadata={
                        "id": file["id"],
                        "name": file["name"],
                        "mime_type": file["mimeType"],
                    },
                )
                documents.append(document)

            return documents

        except httpx.HTTPError as error:
            raise ValueError(f"Error fetching Google Drive files: {error}")

    def _get_file_content(self, file_id: str) -> str:
        """
        Retrieve text content from a Google Drive file.

        :param file_id: ID of the Google Drive file
        :return: Text content of the file
        """
        try:
            # Prepare export request
            headers = {"Authorization": f"Bearer {self.access_token}"}

            # Export file as plain text
            export_url = f"https://www.googleapis.com/drive/v3/files/{file_id}/export"
            params = {"mimeType": "text/plain"}

            response = self.client.get(export_url, headers=headers, params=params)
            response.raise_for_status()

            return response.text

        except httpx.HTTPError as error:
            print(f"An error occurred retrieving file content: {error}")
            return ""

    def insert_data(self, data, **kwargs):
        """
        Insert a new file into Google Drive.

        :param data: Content of the file to be inserted
        :param kwargs: Additional metadata like filename, mime_type
        :return: ID of the inserted file
        """
        if not self.access_token:
            raise ValueError("Not authenticated. Call authenticate() first.")

        try:
            headers = {
                "Authorization": f"Bearer {self.access_token}",
                "Content-Type": "application/json",
            }

            # Prepare file metadata
            file_metadata = {
                "name": kwargs.get("filename", "Untitled Document"),
                "mimeType": kwargs.get(
                    "mime_type", "application/vnd.google-apps.document"
                ),
            }

            # Prepare file content (base64 encoded)
            media_content = base64.b64encode(data.encode("utf-8")).decode("utf-8")

            # Construct payload
            payload = {
                "metadata": file_metadata,
                "media": {"mimeType": "text/plain", "body": media_content},
            }

            # Make request to create file
            response = self.client.post(
                "https://www.googleapis.com/upload/drive/v3/files",
                headers=headers,
                json=payload,
            )
            response.raise_for_status()

            return response.json().get("id")

        except httpx.HTTPError as error:
            raise ValueError(f"Error inserting file: {error}")

    def update_data(self, identifier, data, **kwargs):
        """
        Update an existing Google Drive file.

        :param identifier: File ID to update
        :param data: New content for the file
        :param kwargs: Additional update parameters
        """
        if not self.access_token:
            raise ValueError("Not authenticated. Call authenticate() first.")

        try:
            headers = {
                "Authorization": f"Bearer {self.access_token}",
                "Content-Type": "application/json",
            }

            # Prepare file content (base64 encoded)
            media_content = base64.b64encode(data.encode("utf-8")).decode("utf-8")

            # Construct payload
            payload = {"media": {"mimeType": "text/plain", "body": media_content}}

            # Make request to update file
            response = self.client.patch(
                f"https://www.googleapis.com/upload/drive/v3/files/{identifier}",
                headers=headers,
                json=payload,
            )
            response.raise_for_status()

        except httpx.HTTPError as error:
            raise ValueError(f"Error updating file: {error}")

    def delete_data(self, identifier, **kwargs):
        """
        Delete a file from Google Drive.

        :param identifier: File ID to delete
        """
        if not self.access_token:
            raise ValueError("Not authenticated. Call authenticate() first.")

        try:
            headers = {"Authorization": f"Bearer {self.access_token}"}

            response = self.client.delete(
                f"https://www.googleapis.com/drive/v3/files/{identifier}",
                headers=headers,
            )
            response.raise_for_status()

        except httpx.HTTPError as error:
            raise ValueError(f"Error deleting file: {error}")

    def test_connection(self, **kwargs):
        """
        Test the connection to Google Drive by listing files.

        :return: Boolean indicating connection success
        """
        try:
            if not self.access_token:
                self.authenticate(**kwargs)

            # Prepare request headers
            headers = {
                "Authorization": f"Bearer {self.access_token}",
                "Accept": "application/json",
            }

            # List first 5 files to test connection
            params = {"pageSize": 5, "fields": "files(id,name)"}

            response = self.client.get(
                "https://www.googleapis.com/drive/v3/files",
                headers=headers,
                params=params,
            )
            response.raise_for_status()

            files = response.json().get("files", [])
            return len(files) > 0

        except Exception as e:
            print(f"Connection test failed: {e}")
            return False



from typing import Dict, Any, Literal
from swarmauri_base.tools.ToolBase import ToolBase
from swarmauri_base.schema_converters.SchemaConverterBase import SchemaConverterBase
from swarmauri_core.ComponentBase import ComponentBase, SubclassUnion

@ComponentBase.register_type(SchemaConverterBase, 'CohereSchemaConverter')
class CohereSchemaConverter(SchemaConverterBase):
    type: Literal['CohereSchemaConverter'] = 'CohereSchemaConverter'

    def convert(self, tool: SubclassUnion[ToolBase]) -> Dict[str, Any]:
        properties = {}

        for param in tool.parameters:
            properties[param.name] = {
                "description": param.description,
                "required": param.required
            }
            if param.type == 'string':
                _type = 'str'
            elif param.type == 'float':
                _type = 'float'
            elif param.type == 'integer':
                _type = 'int'
            elif param.type == 'boolean':
                _type = 'bool'
            else:
                raise NotImplementedError(f'🚧 Support for missing type pending https://docs.cohere.com/docs/parameter-types-in-tool-use\n: Missing Type: {param.type}')
            properties[param.name].update({'type': _type})

        return {
            "name": tool.name,
            "description": tool.description,
            "parameter_definitions": properties
        }


from typing import Literal
from swarmauri_base.control_panels.ControlPanelBase import ControlPanelBase
from swarmauri_core.ComponentBase import ComponentBase

@ComponentBase.register_type(ControlPanelBase, 'ControlPanel')
class ControlPanel(ControlPanelBase):
    """
    Concrete implementation of the ControlPanelBase.
    """

    type: Literal["ControlPanel"] = "ControlPanel"



# from swarmauri_standard.utils._lazy_import import _lazy_import

# # List of parsers names (file names without the ".py" extension) and corresponding class names
# parsers_files = [
#     (
#         "swarmauri_standard.parsers.BeautifulSoupElementParser",
#         "BeautifulSoupElementParser",
#     ),
#     ("swarmauri_standard.parsers.CSVParser", "CSVParser"),
#     ("swarmauri_standard.parsers.HTMLTagStripParser", "HTMLTagStripParser"),
#     ("swarmauri_standard.parsers.KeywordExtractorParser", "KeywordExtractorParser"),
#     ("swarmauri_standard.parsers.Md2HtmlParser", "Md2HtmlParser"),
#     ("swarmauri_standard.parsers.OpenAPISpecParser", "OpenAPISpecParser"),
#     (
#         "swarmauri_standard.parsers.PhoneNumberExtractorParser",
#         "PhoneNumberExtractorParser",
#     ),
#     ("swarmauri_standard.parsers.PythonParser", "PythonParser"),
#     ("swarmauri_standard.parsers.RegExParser", "RegExParser"),
#     ("swarmauri_standard.parsers.URLExtractorParser", "URLExtractorParser"),
#     ("swarmauri_standard.parsers.XMLParser", "XMLParser"),
# ]

# # Lazy loading of parsers classes, storing them in variables
# for module_name, class_name in parsers_files:
#     globals()[class_name] = _lazy_import(module_name, class_name)

# # Adding the lazy-loaded parsers classes to __all__
# __all__ = [class_name for _, class_name in parsers_files]



from typing import Dict, Any, List, Optional, Set, Literal
from uuid import uuid4
import asyncio
from swarmauri_base.transports.TransportBase import TransportBase, TransportProtocol
from swarmauri_core.ComponentBase import ComponentBase

@ComponentBase.register_type(TransportBase, 'PubSubTransport')
class PubSubTransport(TransportBase):
    allowed_protocols: List[TransportProtocol] = [TransportProtocol.PUBSUB]
    _topics: Dict[str, Set[str]] = {}  # Topic to subscriber mappings
    _subscribers: Dict[str, asyncio.Queue] = {}
    type: Literal["PubSubTransport"] = "PubSubTransport"

    async def subscribe(self, topic: str) -> str:
        """
        Subscribe an agent to a specific topic.

        Args:
            topic (str): The topic to subscribe to

        Returns:
            str: Unique subscriber ID
        """
        subscriber_id = self.id

        # Create message queue for this subscribere
        self._subscribers[subscriber_id] = asyncio.Queue()

        # Add subscriber to topic
        if topic not in self._topics:
            self._topics[topic] = set()
        self._topics[topic].add(subscriber_id)

        return subscriber_id

    async def unsubscribe(self, topic: str):
        """
        Unsubscribe an agent from a topic.

        Args:
            topic (str): The topic to unsubscribe from
            subscriber_id (str): Unique identifier of the subscriber
        """
        subscriber_id = self.id
        if topic in self._topics and subscriber_id in self._topics[topic]:
            self._topics[topic].remove(subscriber_id)

            # Optional: Clean up if no subscribers remain
            if not self._topics[topic]:
                del self._topics[topic]

    async def publish(self, topic: str, message: Any):
        """
        Publish a message to a specific topic.

        Args:
            topic (str): The topic to publish to
            message (Any): The message to be published
        """
        if topic not in self._topics:
            return

        # Distribute message to all subscribers of this topic
        for subscriber_id in self._topics[topic]:
            await self._subscribers[subscriber_id].put(message)

    async def receive(self) -> Any:
        """
        Receive messages for a specific subscriber.

        Args:
            subscriber_id (str): Unique identifier of the subscriber

        Returns:
            Any: Received message
        """
        return await self._subscribers[self.id].get()

    def send(self, sender: str, recipient: str, message: Any) -> None:
        """
        Simulate sending a direct message (not applicable in Pub/Sub context).

        Args:
            sender (str): The sender ID
            recipient (str): The recipient ID
            message (Any): The message to send

        Raises:
            NotImplementedError: This method is not applicable for Pub/Sub.
        """
        raise NotImplementedError("Direct send not supported in Pub/Sub model.")

    def broadcast(self, sender: str, message: Any) -> None:
        """
        Broadcast a message to all subscribers of all topics.

        Args:
            sender (str): The sender ID
            message (Any): The message to broadcast
        """
        for topic in self._topics:
            asyncio.create_task(self.publish(topic, message))

    def multicast(self, sender: str, recipients: List[str], message: Any) -> None:
        """
        Send a message to specific topics (acting as recipients).

        Args:
            sender (str): The sender ID
            recipients (List[str]): Topics to send the message to
            message (Any): The message to send
        """
        for topic in recipients:
            asyncio.create_task(self.publish(topic, message))


check = PubSubTransport()
print(check.type)
print("I am okay")



from typing import Optional, Union, List, Literal
from pydantic import Field, ConfigDict, field_validator
from swarmauri_standard.messages.SystemMessage import SystemMessage
from swarmauri_standard.messages.HumanMessage import HumanMessage
from swarmauri_standard.messages.AgentMessage import AgentMessage
from swarmauri_standard.exceptions.IndexErrorWithContext import IndexErrorWithContext
from swarmauri_base.conversations.ConversationBase import ConversationBase
from swarmauri_base.conversations.ConversationSystemContextMixin import ConversationSystemContextMixin
from swarmauri_core.conversations.IMaxSize import IMaxSize
from swarmauri_core.messages.IMessage import IMessage
from swarmauri_core.ComponentBase import ComponentBase

@ComponentBase.register_type(ConversationBase, 'MaxSystemContextConversation')
class MaxSystemContextConversation(IMaxSize, ConversationSystemContextMixin, ConversationBase):
    system_context: Optional[SystemMessage] = SystemMessage(content="")
    max_size: int = Field(default=2, gt=1)
    model_config = ConfigDict(extra='forbid', arbitrary_types_allowed=True)
    type: Literal['MaxSystemContextConversation'] = 'MaxSystemContextConversation'
    
    @field_validator('system_context', mode='before')
    def set_system_context(cls, value: Union[str, SystemMessage]) -> SystemMessage:
        if isinstance(value, str):
            return SystemMessage(content=value)
        return value
    
    @property
    def history(self) -> List[IMessage]:
        """
        Get the conversation history, ensuring it starts with a 'user' message and alternates correctly between 'user' and 'assistant' roles.
        The maximum number of messages returned does not exceed max_size + 1.
        """
        res = []  # Start with an empty list to build the proper history

        # Attempt to find the first 'user' message in the history.
        user_start_index = -1
        for index, message in enumerate(self._history):
            if isinstance(message, HumanMessage):  # Identify user message
                user_start_index = index
                break

        # If no 'user' message is found, just return the system context.
        if user_start_index == -1:
            return [self.system_context]

        # Build history from the first 'user' message ensuring alternating roles.
        res.append(self.system_context)
        alternating = True
        count = 0 
        for message in self._history[user_start_index:]:
            if count > self.max_size: # max size
                break
            if alternating and isinstance(message, HumanMessage) or not alternating and isinstance(message, AgentMessage):
                res.append(message)
                alternating = not alternating
                count += 1
            elif not alternating and isinstance(message, HumanMessage):
                # If we find two 'user' messages in a row when expecting an 'assistant' message, we skip this 'user' message.
                continue
            else:
                # If there is no valid alternate message to append, break the loop
                break

        return res

    def add_message(self, message: IMessage):
        """
        Adds a message to the conversation history and ensures history does not exceed the max size.
        """
        if isinstance(message, SystemMessage):
            raise ValueError(f"System context cannot be set through this method on {self.__class_name__}.")
        elif isinstance(message, IMessage):
            self._history.append(message)
        else:
            raise ValueError(f"Must use a subclass of IMessage")
        self._enforce_max_size_limit()
        
    def _enforce_max_size_limit(self):
        """
        Remove messages from the beginning of the conversation history if the limit is exceeded.
        We add one to max_size to account for the system context message
        """
        try:
            while len(self._history) > self.max_size + 1:
                self._history.pop(0)
                self._history.pop(0)
        except IndexError as e:
            raise IndexErrorWithContext(e)



from typing import List, Literal
from swarmauri_standard.vectors.Vector import Vector
from swarmauri_base.distances.DistanceBase import DistanceBase
from swarmauri_core.ComponentBase import ComponentBase

@ComponentBase.register_type(DistanceBase, 'SquaredEuclideanDistance')
class SquaredEuclideanDistance(DistanceBase):
    """
    A concrete class for computing the squared Euclidean distance between two vectors.
    """
    type: Literal['SquaredEuclideanDistance'] = 'SquaredEuclideanDistance'
    
    def distance(self, vector_a: Vector, vector_b: Vector) -> float:
        """
        Computes the squared Euclidean distance between vectors `vector_a` and `vector_b`.

        Parameters:
        - vector_a (Vector): The first vector in the comparison.
        - vector_b (Vector): The second vector in the comparison.

        Returns:
        - float: The computed squared Euclidean distance between vector_a and vector_b.
        """
        if vector_a.shape != vector_b.shape:
            raise ValueError("Vectors must be of the same dimensionality.")

        squared_distance = sum((a - b) ** 2 for a, b in zip(vector_a.value, vector_b.value))
        return squared_distance

    def similarity(self, vector_a: Vector, vector_b: Vector) -> float:
        """
        Squared Euclidean distance is not used for calculating similarity.
        
        Parameters:
        - vector_a (Vector): The first vector.
        - vector_b (Vector): The second vector.

        Raises:
        - NotImplementedError: Indicates that similarity calculation is not implemented.
        """
        raise NotImplementedError("Similarity calculation is not implemented for Squared Euclidean distance.")
        
        
    def distances(self, vector_a: Vector, vectors_b: List[Vector]) -> List[float]:
        distances = [self.distance(vector_a, vector_b) for vector_b in vectors_b]
        return distances
    
    def similarities(self, vector_a: Vector, vectors_b: List[Vector]) -> List[float]:
        raise NotImplementedError("Similarity calculation is not implemented for Squared Euclidean distance.")


# from swarmauri_standard.utils._lazy_import import _lazy_import

# # List of llms names (file names without the ".py" extension) and corresponding class names
# llms_files = [
#     ("swarmauri_standard.llms.BlackForestImgGenModel", "BlackForestImgGenModel"),
#     ("swarmauri_standard.llms.DeepInfraImgGenModel", "DeepInfraImgGenModel"),
#     ("swarmauri_standard.llms.FalAIImgGenModel", "FalAIImgGenModel"),
#     ("swarmauri_standard.llms.HyperbolicImgGenModel", "HyperbolicImgGenModel"),
#     ("swarmauri_standard.llms.OpenAIImgGenModel", "OpenAIImgGenModel"),
# ]

# # Lazy loading of llms classes, storing them in variables
# for module_name, class_name in llms_files:
#     globals()[class_name] = _lazy_import(module_name, class_name)

# # Adding the lazy-loaded llms classes to __all__
# __all__ = [class_name for _, class_name in llms_files]



from typing import List, Literal, Union
from swarmauri_base.tools.ParameterBase import ParameterBase
from swarmauri_standard.tools.Parameter import Parameter
from swarmauri_base.tools.ToolBase import ToolBase
from swarmauri_core.ComponentBase import ComponentBase

@ComponentBase.register_type(ParameterBase, 'Parameter')
class Parameter(ParameterBase):
    type: Union[
        Literal["string", "number", "boolean", "array", "object"],
        str,
        List[str],
    ]



from typing import Literal
from swarmauri_base.prompts.PromptBase import PromptBase
from swarmauri_core.ComponentBase import ComponentBase

@ComponentBase.register_type(PromptBase, 'Prompt')
class Prompt(PromptBase):
    type: Literal['Prompt'] = 'Prompt'


# Import necessary modules
import requests
from typing import Optional, Dict, Any, Literal, List
from pydantic import Field
from swarmauri_standard.tools.Parameter import Parameter
from swarmauri_base.tools.ToolBase import ToolBase
from swarmauri_core.ComponentBase import ComponentBase

@ComponentBase.register_type(ToolBase, 'JSONRequestsTool')
class JSONRequestsTool(ToolBase):
    """
    A tool that leverages the `requests` library to perform HTTP operations.

    Attributes:
        name (str): The name of the tool.
        description (str): A brief description of what the tool does.
    """

    name: str = "JSONRequestsTool"
    type: Literal["JSONRequestsTool"] = "JSONRequestsTool"
    description: Optional[str] = (
        "A tool for making HTTP requests using the `requests` library."
    )
    parameters: List[Parameter] = Field(
        default_factory=lambda: [
            Parameter(
                name="method",
                type="string",
                description="The HTTP method to use ('get', 'post', 'put', 'delete').",
                required=True,
                enum=["get", "post", "put", "delete"],
            ),
            Parameter(
                name="url",
                type="string",
                description="The URL for the request.",
                required=True,
            ),
            Parameter(
                name="params",
                type="object",
                description="The query parameters to include in the request.",
                required=False,
            ),
            Parameter(
                name="data",
                type="object",
                description="The form data to include in the request (used in POST and PUT).",
                required=False,
            ),
            Parameter(
                name="json",
                type="object",
                description="The JSON data to include in the request (used in POST and PUT).",
                required=False,
            ),
            Parameter(
                name="headers",
                type="object",
                description="Additional headers to include in the request.",
                required=False,
            ),
        ]
    )

    def get(
        self,
        url: str,
        params: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, str]] = None,
    ) -> requests.Response:
        """
        Perform an HTTP GET request.

        Args:
            url (str): The URL to send the GET request to.
            params (Optional[Dict[str, Any]]): The query parameters to include in the request.
            headers (Optional[Dict[str, str]]): Additional headers to include in the request.

        Returns:
            requests.Response: The response object from the GET request.
        """
        response = requests.get(url, params=params, headers=headers)
        response.raise_for_status()  # Raise an HTTPError for bad responses (4xx and 5xx)
        return response

    def post(
        self,
        url: str,
        data: Optional[Dict[str, Any]] = None,
        json: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, str]] = None,
    ) -> requests.Response:
        """
        Perform an HTTP POST request.

        Args:
            url (str): The URL to send the POST request to.
            data (Optional[Dict[str, Any]]): The form data to include in the request.
            json (Optional[Dict[str, Any]]): The JSON data to include in the request.
            headers (Optional[Dict[str, str]]): Additional headers to include in the request.

        Returns:
            requests.Response: The response object from the POST request.
        """
        response = requests.post(url, data=data, json=json, headers=headers)
        response.raise_for_status()  # Raise an HTTPError for bad responses (4xx and 5xx)
        return response

    def put(
        self,
        url: str,
        data: Optional[Dict[str, Any]] = None,
        json: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, str]] = None,
    ) -> requests.Response:
        """
        Perform an HTTP PUT request.

        Args:
            url (str): The URL to send the PUT request to.
            data (Optional[Dict[str, Any]]): The form data to include in the request.
            json (Optional[Dict[str, Any]]): The JSON data to include in the request.
            headers (Optional[Dict[str, str]]): Additional headers to include in the request.

        Returns:
            requests.Response: The response object from the PUT request.
        """
        response = requests.put(url, data=data, json=json, headers=headers)
        response.raise_for_status()  # Raise an HTTPError for bad responses (4xx and 5xx)
        return response

    def delete(
        self, url: str, headers: Optional[Dict[str, str]] = None
    ) -> requests.Response:
        """
        Perform an HTTP DELETE request.

        Args:
            url (str): The URL to send the DELETE request to.
            headers (Optional[Dict[str, str]]): Additional headers to include in the request.

        Returns:
            requests.Response: The response object from the DELETE request.
        """
        response = requests.delete(url, headers=headers)
        response.raise_for_status()  # Raise an HTTPError for bad responses (4xx and 5xx)
        return response

    def __call__(self, method: str, url: str, **kwargs) -> Dict[str, Any]:
        """
        Calls the appropriate HTTP method (GET, POST, PUT, DELETE) based on the method argument.

        Args:
            method (str): The HTTP method to use ('get', 'post', 'put', 'delete').
            url (str): The URL for the request.
            **kwargs: Additional keyword arguments passed to the respective method.

        Returns:
            requests.Response: The response object from the HTTP request.
        """
        method = method.lower()
        if method == "get":
            response = self.get(url, **kwargs)
        elif method == "post":
            response = self.post(url, **kwargs)
        elif method == "put":
            response = self.put(url, **kwargs)
        elif method == "delete":
            response = self.delete(url, **kwargs)
        else:
            raise ValueError(f"Unsupported HTTP method: {method}")

        return response.json()




from typing import Literal
from swarmauri_base.toolkits.ToolkitBase import ToolkitBase
from swarmauri_core.ComponentBase import ComponentBase

@ComponentBase.register_type(ToolkitBase, 'Toolkit')
class Toolkit(ToolkitBase):
    type: Literal['Toolkit'] = 'Toolkit'


import sys
import io
from typing import List, Literal, Dict
from pydantic import Field
from swarmauri_standard.tools.Parameter import Parameter
from swarmauri_base.tools.ToolBase import ToolBase
from swarmauri_core.ComponentBase import ComponentBase

@ComponentBase.register_type(ToolBase, 'CodeInterpreterTool')
class CodeInterpreterTool(ToolBase):
    version: str = "1.0.0"
    parameters: List[Parameter] = Field(
        default_factory=lambda: [
            Parameter(
                name="user_code",
                type="string",
                description=(
                    "Executes the provided Python code snippet in a secure sandbox environment. "
                    "This tool is designed to interpret the execution of the python code snippet."
                    "Returns the output"
                ),
                required=True,
            )
        ]
    )
    name: str = "CodeInterpreterTool"
    description: str = "Executes provided Python code and captures its output."
    type: Literal["CodeInterpreterTool"] = "CodeInterpreterTool"

    def __call__(self, user_code: str) -> Dict[str, str]:
        """
        Executes the provided user code and captures its stdout output.

        Parameters:
            user_code (str): Python code to be executed.

        Returns:
            str: Captured output or error message from the executed code.
        """
        return {"code_output": self.execute_code(user_code)}

    def execute_code(self, user_code: str) -> str:
        """
        Executes the provided user code and captures its stdout output.

        Args:
            user_code (str): Python code to be executed.

        Returns:
            str: Captured output or error message from the executed code.
        """
        old_stdout = sys.stdout
        redirected_output = sys.stdout = io.StringIO()

        try:
            # Note: Consider security implications of using 'exec'
            builtins = globals().copy()
            exec(user_code, builtins)
            sys.stdout = old_stdout
            captured_output = redirected_output.getvalue()
            return str(captured_output)
        except Exception as e:
            sys.stdout = old_stdout
            return f"An error occurred: {str(e)}"


from typing import Literal
from swarmauri_base.chains.PromptContextChainBase import PromptContextChainBase
from swarmauri_core.ComponentBase import ComponentBase

@ComponentBase.register_type(PromptContextChainBase, 'PromptContextChain')
class PromptContextChain(PromptContextChainBase):
    type: Literal['PromptContextChain'] = 'PromptContextChain'


import re
from typing import List, Union, Any, Literal
from swarmauri_standard.documents.Document import Document
from swarmauri_base.parsers.ParserBase import ParserBase
from swarmauri_core.ComponentBase import ComponentBase

@ComponentBase.register_type(ParserBase, 'PhoneNumberExtractorParser')
class PhoneNumberExtractorParser(ParserBase):
    """
    A parser that extracts phone numbers from the input text.
    Utilizes regular expressions to identify phone numbers in various formats.
    """
    type: Literal['PhoneNumberExtractorParser'] = 'PhoneNumberExtractorParser'
    
    def parse(self, data: Union[str, Any]) -> List[Document]:
        """
        Parses the input data, looking for phone numbers employing a regular expression.
        Each phone number found is contained in a separate IDocument instance.

        Parameters:
        - data (Union[str, Any]): The input text to be parsed for phone numbers.

        Returns:
        - List[IDocument]: A list of IDocument instances, each containing a phone number.
        """
        # Define a regular expression for phone numbers.
        # This is a simple example and might not capture all phone number formats accurately.
        phone_regex = r'\+?\d[\d -]{8,}\d'

        # Find all occurrences of phone numbers in the text
        phone_numbers = re.findall(phone_regex, str(data))

        # Create a new IDocument for each phone number found
        documents = [Document(content=phone_number, metadata={}) for index, phone_number in enumerate(phone_numbers)]

        return documents


from typing import List, Literal, Dict
import nltk
from nltk.tokenize import sent_tokenize, word_tokenize
from pydantic import Field
from swarmauri_standard.tools.Parameter import Parameter
from swarmauri_base.tools.ToolBase import ToolBase
from swarmauri_core.ComponentBase import ComponentBase

# Download required NLTK data once during module load
nltk.download("punkt", quiet=True)


@ComponentBase.register_type(ToolBase, 'SentenceComplexityTool')
class SentenceComplexityTool(ToolBase):
    version: str = "0.1.dev2"
    parameters: List[Parameter] = Field(
        default_factory=lambda: [
            Parameter(
                name="text",
                type="string",
                description="The text to analyze for sentence complexity.",
                required=True,
            )
        ]
    )

    name: str = "SentenceComplexityTool"
    description: str = "Evaluates sentence complexity based on average sentence length and the number of clauses."
    type: Literal["SentenceComplexityTool"] = "SentenceComplexityTool"

    def __call__(self, text: str) -> Dict[str, float]:
        """
        Evaluate sentence complexity based on average sentence length and the number of clauses.

        Parameters:
        - text (str): The text to analyze.

        Returns:
        - dict: A dictionary containing average sentence length and average number of clauses per sentence.
        """
        if not text.strip():
            raise ValueError("Input text cannot be empty.")

        sentences = sent_tokenize(text)
        num_sentences = len(sentences)

        if num_sentences == 0:
            return {"average_sentence_length": 0.0, "average_clauses_per_sentence": 0.0}

        total_words = 0
        total_clauses = 0

        for sentence in sentences:
            words = word_tokenize(sentence)
            total_words += len(words)

            # Improved clause counting method
            clauses = sentence.count(",") + sentence.count(";")
            clauses += sum(
                sentence.lower().count(conj)
                for conj in [
                    "and",
                    "but",
                    "or",
                    "because",
                    "although",
                    "though",
                    "while",
                    "if",
                ]
            )
            total_clauses += clauses + 1

        avg_sentence_length = total_words / num_sentences
        avg_clauses_per_sentence = total_clauses / num_sentences

        return {
            "average_sentence_length": avg_sentence_length,
            "average_clauses_per_sentence": avg_clauses_per_sentence,
        }




from typing import List, Literal, Dict
from pydantic import Field
from swarmauri_standard.tools.Parameter import Parameter
from swarmauri_base.tools.ToolBase import ToolBase
from swarmauri_core.ComponentBase import ComponentBase

@ComponentBase.register_type(ToolBase, 'WeatherTool')
class WeatherTool(ToolBase):
    version: str = "0.1.dev1"
    parameters: List[Parameter] = Field(
        default_factory=lambda: [
            Parameter(
                name="location",
                type="string",
                description="The location for which to fetch weather information",
                required=True,
            ),
            Parameter(
                name="unit",
                type="string",
                description="The unit for temperature ('fahrenheit' or 'celsius')",
                required=True,
                enum=["fahrenheit", "celsius"],
            ),
        ]
    )
    name: str = "WeatherTool"
    description: str = "Fetch current weather info for a location"
    type: Literal["WeatherTool"] = "WeatherTool"

    def __call__(self, location, unit="fahrenheit") -> Dict[str, str]:
        weather_info = (location, unit)
        # Here you would implement the actual logic for fetching the weather information.
        # For demonstration, let's just return the parameters as a string.
        return {"weather_info": str(weather_info)}

